/*                               -*- Mode: C -*- 
 * loop.c --- Streams loopback driver, as of Sun manual 
 * Author          : Graham Wheeler
 * Created On      : Sat Oct  7 05:01:31 1995
 * Last Modified By: David Grothe
 * RCS Id          : $Id: loop.c,v 1.5 1996/01/24 19:44:35 dave Exp $
 * Purpose         : provide loopback streams driver
 * ----------------______________________________________________
 *
 *    Copyright (C) 1995  Graham Wheeler, Francisco J. Ballesteros,
 *    Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach us by email to any of
 *    gram@aztec.co.za, nemo@ordago.uc3m.es
 */

#ident "@(#) LiS loop.c 2.4 6/17/99 14:46:55 "

/*  -------------------------------------------------------------------  */

#ifdef MODULE
#  if defined(LINUX) && defined(__KERNEL__)
#    ifdef MODVERSIONS
#      include <linux/modversions.h>
#    endif
#    include <linux/module.h>
#  else
#    error This can only be a module in the Linux kernel environment
#  endif
#  include <sys/LiS/config.h>
#else
#  define MOD_INC_USE_COUNT
#  define MOD_DEC_USE_COUNT
#endif

#include <sys/stream.h>
#include <sys/stropts.h>

/*  -------------------------------------------------------------------  */
/*			  Module definition structs                      */

/* Module info for the loopback driver
 */
static struct module_info loop_minfo =
{
  0,				/* id */
  "loop",			/* name */
  0,				/* min packet size accepted */
  INFPSZ,			/* max packet size accepted */
  10240L,			/* high water mark */
  512L				/* low water mark */
};

/* These are the entry points to the driver: open, close, write side put and
 * service procedures and read side service procedure.
 */
static int   loop_open  (queue_t *,dev_t*,int,int, cred_t *);
static int   loop_close (queue_t *, int, cred_t *);
static void  loop_wput (queue_t *, mblk_t *);
static void  loop_wsrv  (queue_t *);
static void  loop_rsrv  (queue_t *);

/* qinit structures (rd and wr side) 
 */
static struct qinit loop_rinit =
{
  NULL,				/* put */       
  loop_rsrv,			/* service  */  
  loop_open,			/* open */      
  loop_close,			/* close */     
  NULL,				/* admin */     
  &loop_minfo,			/* info */      
  NULL				/* stat */      
};

static struct qinit loop_winit =
{
  loop_wput,                    /* put */       
  loop_wsrv, 			/* service  */  
  NULL, 			/* open */      
  NULL, 			/* close */     
  NULL, 			/* admin */     
  &loop_minfo, 			/* info */      
  NULL				/* stat */      
};

/* streamtab for the loopback driver.
 */
struct streamtab sloop_info =
{
  &loop_rinit,			/* read queue */
  &loop_winit,			/* write queue */
  NULL,				/* mux read queue  */
  NULL				/* mux write queue */
};

/*  -------------------------------------------------------------------  */
/*			    Module implementation                        */

#include <sys/LiS/loop.h>

/*  -------------------------------------------------------------------  */

#define NLOOP 8			/* number of clone devs supported */

struct loop			/* info for each loopback device */
{
  queue_t *qptr;
  queue_t *oqptr;
  int	   use_putnext ;	/* instead of service queue */
  int	   use_bufcall ;	/* instead of putnext */
  int	   msglvl ;		/* # msgs to queue before sending upstream */
  int	   msgcnt ;		/* # msgs in the queue */
  int	   timr ;		/* time out amount while msgs queued */
  int	   timr_hndl ;		/* for running timer */
  int	   mark ;		/* "mark" the next message(s) from above */
  int	   minor_nr ;		/* minor number of this instance */
  int	   catlvl ;		/* # msgs to concat before sending upstream */
  int	   copy_bfr ;		/* copy the message before looping */
  mblk_t  *saved_msg ;		/* msg saved for contatenation */
  char	  *copy_buf ;		/* for copyin/out */
  int      copy_buf_size ;	/* length of copy_buf data */
  loop_xparent_t xp ;		/* copied in from user space */
};

/* Arry.  w/ opened devices info & number of openened devices
 */
static struct loop loop_loop[NLOOP];
static int loop_cnt = NLOOP;

/*  -------------------------------------------------------------------  */

void	sloop_init(void)
{
    printk("Loop-around driver init routine called\n") ;

} /* loop_init */

/*  -------------------------------------------------------------------  */


static int
loop_open (queue_t *q, dev_t *devp, int flag, int sflag, cred_t *credp)
{
  struct loop *loop;
  dev_t		dev ;

  if (sflag == CLONEOPEN) {
    for (dev = 1; dev < loop_cnt; dev++) {
      if (loop_loop[dev].qptr == NULL)
	break;
    }
  } else
    dev = minor (*devp);

  if (dev >= loop_cnt)
    return OPENFAIL;

  *devp = makedevice(major(*devp), dev) ;
  if (q->q_ptr)
    return 0;					/* success */

  loop = &loop_loop[dev];
  WR (q)->q_ptr = (char *) loop;
  q->q_ptr = (char *) loop;
  loop->qptr = WR (q);
  loop->minor_nr = dev ;

#ifdef STREAMS_DEBUG
  printk("Opened %x\n",dev);
#endif

  MOD_INC_USE_COUNT;

  return 0;					/* success */
}

/*  -------------------------------------------------------------------  */

static void	loop_timeout(caddr_t arg)
{
    struct loop		*loop = (struct loop *) arg ;

    if (loop->qptr != NULL)
    {
	enableok(WR(loop->qptr)) ;	/* allow qenable to work */
	qenable(WR(loop->qptr)) ;
    }

} /* loop_timeout */

/*  -------------------------------------------------------------------  */
/*
 * loop_bufcall
 *
 * Called from the STREAMS bufcall mechanism.
 *
 * The argument is a pointer to a queue to enable.
 */
static void loop_bufcall(long q_ptr)
{
    enableok((queue_t *) q_ptr) ;	/* allow qenable to work */
    qenable((queue_t *) q_ptr) ;

} /* loop_bufcall */

/*  -------------------------------------------------------------------  */
/*
 * Concatenate the given message with the "saved" message in the
 * loop structure.  This allows the user to build multi-segment
 * messages that later get sent upstream.
 */
static mblk_t *loop_concat(struct loop *loop, mblk_t *mp)
{
    mblk_t	*lastmp ;

    if (loop->saved_msg == NULL)
    {
	loop->saved_msg = mp ;
	return(mp) ;
    }

    for (lastmp = loop->saved_msg;
	 lastmp->b_cont != NULL;
	 lastmp = lastmp->b_cont)
    ;
	
    lastmp->b_cont = mp ;		/* hook new onto saved msg */
    mp = loop->saved_msg ;		/* mp becomes new concatenated msg */
    loop->saved_msg = NULL ;		/* no longer saved */
    return(mp) ;			/* return entire message */

} /* loop_concat */

/*  -------------------------------------------------------------------  */

int loop_request_copyio(queue_t *wq, struct loop *loop, mblk_t *mp,
			int msg_type, void *uptr, int nbytes)
{
    struct copyreq	*rq = (struct copyreq *) mp->b_rptr ;
    mblk_t		*dp = mp->b_cont ;

    mp->b_wptr = mp->b_rptr + sizeof(*rq) ;
    rq->cq_addr = uptr ;
    rq->cq_size = nbytes ;
    rq->cq_flag = 0 ;
    rq->cq_private = (mblk_t *) loop ;

    if (dp != NULL && msg_type == M_COPYIN)
    {
	mp->b_cont = NULL ;		/* discard data */
	freemsg(dp) ;
    }

    mp->b_datap->db_type = msg_type ;	/* M_COPYIN or M_COPYOUT */
    qreply(wq, mp) ;
    return(0) ;

} /* loop_request_copyio */


/*  -------------------------------------------------------------------  */

int loop_iocdata(queue_t *wq, mblk_t *mp)
{
    struct copyresp	*rsp = (struct copyresp *) mp->b_rptr ;
    struct loop		*loop = (struct loop *) rsp->cp_private ;
    struct iocblk	*iocb;

    iocb = (struct iocblk *) rsp ;
    switch (rsp->cp_cmd)
    {
    case LOOP_XPARENT_COPYIN:
    case LOOP_XPARENT_COPYOUT:
      {
	mblk_t		*dp ;

	dp = mp->b_cont ;
	if (rsp->cp_rval != 0)
	{
	    printk("loop_iocdata: copyin request failed\n") ;
	    return(-EINVAL) ;
	}

	if (dp == NULL || dp->b_datap->db_type != M_DATA)
	{
	    printk("loop_iocdata: attached bfr NULL or not M_DATA\n") ;
	    return(-EINVAL) ;
	}

	if (msgdsize(dp) != sizeof(loop_xparent_t))
	{
	    printk("loop_iocdata: expected %d bytes, got %d\n",
		    sizeof(loop_xparent_t), msgdsize(dp)) ;
	    return(-EINVAL) ;
	}

	loop->xp = *( (loop_xparent_t *) dp->b_rptr ) ;
	mp->b_cont = NULL ;
	freemsg(dp) ;

	/*
	 * If we fetched the xparent struct for a copy in then issue
	 * another copy in for the actual data from user space.
	 * If it was for a copy out then return the requested data
	 * in an M_DATA and request it to be copied out to the user's
	 * space.
	 *
	 * Alter the command by oring in an 0xf00 so that when the
	 * iocdata for these requests comes down we can tell that
	 * they are for the final copy functions.  Handled in separate
	 * cases below.
	 */
	switch (rsp->cp_cmd)
	{
	case LOOP_XPARENT_COPYIN:
	    rsp->cp_cmd |= 0xf00 ;		/* alter the command */
	    if (loop_request_copyio(wq, loop, mp, M_COPYIN,
				    loop->xp.p_arg, loop->xp.i_arg) < 0)
	    {
		printk("loop_iocdata: loop_request_copyio failed\n") ;
		return(-EINVAL) ;
	    }
	    return(1) ;			/* keep exchanging msgs w/strm hd */

	case LOOP_XPARENT_COPYOUT:
	    if (loop->copy_buf == NULL)
		return(-EINVAL) ;

	    if (loop->copy_buf_size < loop->xp.i_arg)
		loop->xp.i_arg = loop->copy_buf_size ;

	    dp = allocb(loop->xp.i_arg, BPRI_LO) ;
	    if (dp == NULL)
		return(-ENOMEM) ;


	    memcpy(dp->b_wptr, loop->copy_buf, loop->xp.i_arg) ;
	    dp->b_wptr += loop->xp.i_arg ;
	    linkb(mp, dp) ;
	    rsp->cp_cmd |= 0xf00 ;		/* alter the command */
	    if (loop_request_copyio(wq, loop, mp, M_COPYOUT,
				    loop->xp.p_arg, loop->xp.i_arg) < 0)
	    {
		printk("loop_iocdata: loop_request_copyio failed\n") ;
		return(-EINVAL) ;
	    }
	    return(1) ;			/* keep exchanging msgs w/strm hd */
	}
        break ;
      }

    /*
     * This is the actual data that the user wanted us to copy in.
     */
    case LOOP_XPARENT_COPYIN | 0xf00 :
      {
	mblk_t		*dp ;

	dp = mp->b_cont ;
	if (rsp->cp_rval != 0)
	{
	    printk("loop_iocdata: copyin request failed\n") ;
	    return(-EINVAL) ;
	}

	if (dp == NULL || dp->b_datap->db_type != M_DATA)
	{
	    printk("loop_iocdata: attached bfr NULL or not M_DATA\n") ;
	    return(-EINVAL) ;
	}

	if (msgdsize(dp) != loop->xp.i_arg)
	{
	    printk("loop_iocdata: expected %d bytes, got %d\n",
		    loop->xp.i_arg, msgdsize(dp)) ;
	    return(-1) ;
	}

	if (loop->copy_buf != NULL)
	    FREE(loop->copy_buf) ;

	loop->copy_buf = MALLOC(loop->xp.i_arg) ;
	if (loop->copy_buf == NULL)
	    return(-ENOMEM) ;

	memcpy(loop->copy_buf, dp->b_rptr, loop->xp.i_arg) ;
	loop->copy_buf_size = loop->xp.i_arg ;
	return(0) ;			/* send iocack now */
      }

    /*
     * This is the result of copying out data to the user.
     */
    case LOOP_XPARENT_COPYOUT | 0xf00 :
      {
	if (rsp->cp_rval != 0)
	{
	    printk("loop_iocdata: copyin request failed\n") ;
	    return(-EINVAL) ;
	}

	return(0) ;			/* send iocack now */
      }

    default:
	printk("loop_iocdata: cmd=0x%x unknown\n", rsp->cp_cmd) ;
	return(-EINVAL) ;
    }

    return(0) ;

} /* loop_iocdata */

/*  -------------------------------------------------------------------  */

static void
loop_wput (queue_t *q, mblk_t *mp)
{
  struct loop *loop;
  int		rtn_count = 0 ;

  loop = (struct loop *) q->q_ptr;

  switch (mp->b_datap->db_type) {
  case M_IOCTL:{
    struct iocblk *iocb;
    int error;

    iocb = (struct iocblk *) mp->b_rptr;

    switch (iocb->ioc_cmd) {
    case LOOP_SET:{
      int to;

      if (iocb->ioc_count != sizeof (int)) {
	printk ("Expected ioctl len %d, got %d\n", sizeof (int), iocb->ioc_count);

	error = EINVAL;
	goto iocnak;
      }
      to = *(int *) mp->b_cont->b_rptr;

      if (to >= loop_cnt || to < 0 ||
	  !loop_loop[to].qptr) {
	error = ENXIO;
	goto iocnak;
      }
      if (loop->oqptr || loop_loop[to].oqptr) {
	error = EBUSY;
	goto iocnak;
      }
      loop->oqptr = RD (loop_loop[to].qptr);
      loop_loop[to].oqptr = RD (q);
      break;
    }

    case LOOP_PUTNXT:
        loop->use_putnext = 1 ;		/* instead of service queue */
	break ;

    case LOOP_MSGLVL:
	if (mp->b_cont != NULL)
	    loop->msglvl = *(int *) mp->b_cont->b_rptr;
	else
	{
	    printk("loop_wput: no data buffer in ioctl\n") ;
	    error = EINVAL;
	    goto iocnak;
	}
	break ;

    case LOOP_CONCAT:
	if (mp->b_cont != NULL)
	    loop->catlvl = *(int *) mp->b_cont->b_rptr;
	else
	{
	    printk("loop_wput: no data buffer in ioctl\n") ;
	    error = EINVAL;
	    goto iocnak;
	}
	break ;

    case LOOP_TIMR:
	if (mp->b_cont != NULL)
	    loop->timr = *(int *) mp->b_cont->b_rptr;
	else
	{
	    printk("loop_wput: no data buffer in ioctl\n") ;
	    error = EINVAL;
	    goto iocnak;
	}
	break ;

    case LOOP_MARK:
	loop->mark++ ;				/* mark next message */
	break ;

    case LOOP_GET_DEV:				/* get minor number */
	{
	    int		*devp ;

	    if (iocb->ioc_count != sizeof (int))
	    {
		printk ("Expected ioctl len %d, got %d\n",
			 sizeof (int), iocb->ioc_count);

		error = EINVAL;
		goto iocnak;
	    }

	    if (mp->b_cont != NULL)
		devp = (int *) mp->b_cont->b_rptr ;
	    else
	    {
		printk("loop_wput: no data buffer in ioctl\n") ;
		error = EINVAL;
		goto iocnak;
	    }
	    *devp = loop->minor_nr ;		/* return minor device nr */
	    rtn_count = sizeof(int) ;
	}
	break ;

    case LOOP_BUFCALL:
	loop->use_bufcall++ ;
	break ;

    case LOOP_COPY:
	loop->copy_bfr = 1 ;
	break ;

    /*
     * Fetch the loop module's xparent structure from user space.
     * When it arrives then copy data into or out of kernel space
     * as requested.
     */
    case LOOP_XPARENT_COPYIN:
    case LOOP_XPARENT_COPYOUT:
      {
	void		**ptr_ptr ;
	void		*uptr ;

	if (mp->b_cont == NULL)
	{
	    printk("loop_wput: LOOP_XPARENT_COPYIN/OUT: no M_DATA\n") ;
	    error = EINVAL;
	    goto iocnak;
	}

	ptr_ptr = (void **) mp->b_cont->b_rptr ;
	uptr = *ptr_ptr ;
	if (loop_request_copyio(q, loop, mp, M_COPYIN,
				uptr, sizeof(loop_xparent_t)) < 0)
	{
	    error = EINVAL;
	    goto iocnak;
	}
	return ;		/* don't respond w/iocack yet */
      }

    default:
      printk ("Expected ioctl %x, got %x\n", LOOP_SET, iocb->ioc_cmd);
      error = EINVAL;
    iocnak:{
	mp->b_datap->db_type = M_IOCNAK;
	iocb->ioc_error = error;
	qreply (q, mp);
      }
      return ;
    }				/* end of switch on ioc_cmd */

    mp->b_datap->db_type = M_IOCACK;
    iocb->ioc_count = rtn_count;
    qreply (q, mp);

    break;
  }

  case M_IOCDATA:
  {
    struct iocblk *iocb;
    int		   err ;

    iocb = (struct iocblk *) mp->b_rptr;
    if ((err = loop_iocdata(q, mp)) < 0)
    {
	mp->b_datap->db_type = M_IOCNAK;
	iocb->ioc_cmd &= 0xff ;			/* original cmnd */
	iocb->ioc_count = 0;
	iocb->ioc_error = -err;
	mp->b_wptr = mp->b_rptr + sizeof(*iocb) ;
	qreply (q, mp);
	return ;
    }

    if (err > 0)				/* more message xchg going on */
	break ;

    mp->b_datap->db_type = M_IOCACK;
    iocb->ioc_cmd &= 0xff ;			/* original cmnd */
    iocb->ioc_count = 0;
    iocb->ioc_error = 0;
    mp->b_wptr = mp->b_rptr + sizeof(*iocb) ;
    qreply (q, mp);		/* finally ack the ioctl */
    break ;
  }

  case M_FLUSH:{
    if (*mp->b_rptr & FLUSHW)
      flushq (q, FLUSHDATA);
    if (*mp->b_rptr & FLUSHR) {
      flushq (RD (q), FLUSHDATA);
      *mp->b_rptr &= ~FLUSHW;
      qreply (q, mp);
    } else
      freemsg (mp);
    break;
  }
  default:
    if (loop->oqptr == NULL) {
      putctl1 (RD (q)->q_next, M_ERROR, ENXIO);
      freemsg (mp);
      break;
    }

    if (loop->copy_bfr)
    {
	mblk_t	*new_mp ;

	new_mp = copymsg(mp) ;
	if (new_mp == NULL)
	    printk("loop_wput: can't copy message.  Sending original\n") ;
	else
	{
	    freemsg(mp) ;
	    mp = new_mp ;
	}
    }

    if (loop->mark)
    {
	loop->mark-- ;
	mp->b_flag |= MSGMARK ;
    }

    if (loop->catlvl > 0)			/* concatenating msgs */
    {
	mp = loop_concat(loop, mp) ;		/* do it */
	if (--loop->catlvl > 0)			/* send it yet?  */
	{
	    loop->saved_msg = mp ;		/* no, save it */
	    break ;				/* and exit */
	}
    }

    if (loop->use_putnext)
	putnext (loop->oqptr, mp);
    else
    if (loop->use_bufcall)
    {
	noenable(q) ;			/* prevent queue enable */
	putq (q, mp);
	bufcall(0, 0, loop_bufcall, (long) q) ;
    }
    else
    {
	loop->msgcnt++ ;		/* count enqueued message */
	if (loop->msglvl > 0)
	    noenable(q) ;		/* prevent queue enable */

	putq (q, mp);
	if (loop->msgcnt > loop->msglvl)
	    qenable(q) ;
	else
	if (loop->timr != 0 && loop->timr_hndl == 0)
	    loop->timr_hndl = timeout(loop_timeout, (caddr_t) loop, loop->timr) ;
    }
    break;
  
  }
}

/*  -------------------------------------------------------------------  */

static void
loop_wsrv (queue_t *q)
{
  mblk_t *mp;
  register struct loop *loop;

  loop = (struct loop *) q->q_ptr;

  if (   loop->use_bufcall == 0
      && loop->timr_hndl == 0
      && loop->msgcnt <= loop->msglvl) return ;

  if (loop->timr_hndl != 0)
  {
      untimeout(loop->timr_hndl) ;
      loop->timr_hndl = 0 ;
  }

  while ((mp = getq (q)) != NULL)
  {
    if (mp->b_datap->db_type <= QPCTL &&
	!canput (loop->oqptr->q_next))
    {
      putbq (q, mp);
      if (loop->use_bufcall)
	bufcall(0, 0, loop_bufcall, (long) q) ;
      break;
    }

    if (loop->use_bufcall)
	loop->use_bufcall-- ;

    putnext (loop->oqptr, mp);
    if (--loop->msgcnt <= loop->msglvl) break ;
  }
}

/*  -------------------------------------------------------------------  */

static void
loop_rsrv (queue_t *q)
{
  struct loop *loop;

  loop = (struct loop *) q->q_ptr;
  if (loop->oqptr == NULL)
    return;

  qenable (WR (loop->oqptr));
}

/*  -------------------------------------------------------------------  */

static int
loop_close (queue_t *q, int dummy, cred_t *credp)
{
  struct loop *loop;

  loop = (struct loop *) q->q_ptr;
  loop->qptr = NULL;

  if (loop->timr_hndl != 0)
  {
      untimeout(loop->timr_hndl) ;
      loop->timr_hndl = 0 ;
  }

  if (loop->oqptr) {
    ((struct loop *) loop->oqptr->q_ptr)->qptr = NULL;
    ((struct loop *) loop->oqptr->q_ptr)->oqptr = NULL;
    putctl (loop->oqptr->q_next, M_HANGUP);
    loop->oqptr = NULL;
  }
  flushq (q, FLUSHALL);
  flushq (WR (q), FLUSHALL);
  if (loop->saved_msg != NULL)
  {
    freemsg(loop->saved_msg) ;
    loop->saved_msg = NULL ;
  }

  if (loop->copy_buf != NULL)
  {
      FREE(loop->copy_buf) ;
      loop->copy_buf = NULL ;
  }

  loop->use_putnext = 0 ;
  loop->use_bufcall = 0 ;
  loop->msglvl = 0 ;
  loop->msgcnt = 0 ;
  loop->timr = 0 ;
  loop->mark = 0 ;
  loop->copy_bfr = 0 ;
  loop->minor_nr = 0 ;
  loop->catlvl = 0 ;
  MOD_DEC_USE_COUNT;
  return(0);
}


#ifdef MODULE

int init_module(void)
{
        int ret = lis_register_strdev(SLOOP__CMAJOR_0, &sloop_info,
				      NLOOP, "loop");
	if (ret < 0) {
                printk("loop.init_module: Unable to register driver.\n");
                return ret;
        }
	return 0;
}

void cleanup_module(void)
{
        if (lis_unregister_strdev(SLOOP__CMAJOR_0) < 0)
		printk("loop.cleanup_module: Unable to unregister driver.\n");
	else
		printk("loop.cleanup_module: Unregistered, ready to be unloaded.\n");
	return;
}

#endif
