/************************************************************************
*                      SVR4 Driver-Kernel Interface                     *
*************************************************************************
*									*
* These routines implement an SVR4 compatible DKI.			*
*									*
* Author:	David Grothe <dave@gcom.com>				*
*									*
* Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>		*
*									*
************************************************************************/
/*
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 */

#ident "@(#) LiS dki.c 2.1 4/4/99 11:34:23 "

#include <sys/stream.h>

/************************************************************************
*                       SVR4 Compatible timeout                         *
*************************************************************************
*									*
* This implementation of SVR4 compatible timeout functions uses the	*
* Linux timeout mechanism as the internals.				*
*									*
* The technique is to allocate memory that contains a Linux timer	*
* structure and an SVR4-like handle.  We build up a list of these	*
* structures in a linked-list headed by 'lis_tlist_head'.  We never	*
* use handle zero, so that value becomes a marker for an available	*
* entry.  We recycle the list elements as they expire or get untimeout-	*
* ed.									*
*									*
************************************************************************/

typedef struct tlist
{
    struct tlist	*next ;		/* thread of these */
    struct timer_list	 tl ;		/* Linux timer structure */
    int			 handle ;	/* SVR4-style handle */

} tlist_t ;

tlist_t		*lis_tlist_head ;	/* this list is no particular order */
int		 lis_tlist_handle ;	/* next handle to use */

/*
 * "timeout" is #defined to be this function in dki.h.
 */
int	lis_timeout_fcn(timo_fcn_t *timo_fcn, caddr_t arg, long ticks,
		    char *file_name, int line_nr)
{
    tlist_t	*tp ;
    tlist_t	*found ;
    int		 psw ;
    int		 handle ;
    int		 unique ;

    unique = 0 ;
    found = NULL ;
    while (!unique)
    {
	SPLSTR(psw) ;				/* protect handle assgnmnt */
	handle = ++lis_tlist_handle ;		/* next handle */
	if (handle == 0)			/* prevent use of handle 0 */
	    handle = ++lis_tlist_handle ;	/* next handle */
	SPLX(psw) ;

	unique = 1 ;				/* hopefully unique handle */
	for (tp = lis_tlist_head; tp != NULL; tp = tp->next)
	{						/* find one not-in use */
	    if (tp->handle == handle) unique = 0 ;	/* handle not unique */
	    SPLSTR(psw) ;				/* protect finding entry */
	    if (   found == NULL
		&& (   tp->handle == 0
		    || (tp->tl.next == NULL && tp->tl.prev == NULL)
		   )
	       )
	   {
		tp->handle  = 0xFFFF ;		/* protect from "find" */
		tp->tl.next = tp->tl.prev = &tp->tl ;
		found = tp ;
	   }
	   SPLX(psw) ;
	}					/* traverse entire list */
    }

    if (found == NULL)				/* must allocate a new one */
    {
	tp = (tlist_t *) LISALLOC(sizeof(*tp),file_name, line_nr) ;
	if (tp == NULL) return(0) ;		/* no memory for timer */

	SPLSTR(psw) ;				/* protect list insertion */
	tp->next = lis_tlist_head ;
	lis_tlist_head = tp ;
	SPLX(psw) ;
    }
    else
    {
	tp = found ;				/* use found timer */
	lis_mark_mem(tp, file_name, line_nr) ;
    }

    tp->handle = handle ;
    lis_tmout(&tp->tl, (tmout_fcn_t *) timo_fcn, (long) arg, ticks) ;
						/* use Linux */

    return(tp->handle) ;			/* return handle */

} /* lis_timeout_fcn */

int	untimeout(int id)
{
    tlist_t	*tp ;
    int		 psw ;

    for (tp = lis_tlist_head; tp != NULL; tp = tp->next)
    {						/* find one not-in use */
	if (tp->handle == id)			/* is this our timer?  */
	{
	    SPLSTR(psw) ;			/* protect untimout */
	    lis_mark_mem(tp, "unused", MEM_TIMER) ;
	    if (tp->tl.next != NULL || tp->tl.prev != NULL)
		lis_untmout(&tp->tl) ;		/* stop it */

	    tp->handle = 0 ;			/* make available */
	    SPLX(psw) ;
	    break ;				/* done */
	}
    }

    return(0);					/* no rtn value specified */

} /* untimeout */

void lis_terminate_dki(void)
{
	int psw;

	/*
	 *  In case of buggy drivers, timeouts could still happen.
	 *  Better be on the safe side.
	 */
	SPLSTR(psw);
	while (lis_tlist_head != NULL) {
		tlist_t *tp = lis_tlist_head;

		if (tp->tl.next != NULL || tp->tl.prev != NULL) {
			printk("Cancelling forgotten timeout.\n");
			lis_untmout(&tp->tl);
		}
		lis_tlist_head = tp->next;
		FREE(tp);
		SPLX(psw);
		/* Brief interrupt window */
		SPLSTR(psw);
	}
	SPLX(psw);
}

