/*                               -*- Mode: C -*- 
 * msg.c --- streams message handling
 * Author          : Graham Wheeler
 * Created On      : Tue May 31 22:25:19 1994
 * Last Modified By: David Grothe
 * RCS Id          : $Id: msg.c,v 1.6 1996/01/17 18:57:05 dave Exp $
 * Purpose         : provide streams message handling
 *                 : this is the main memory allocation related module.
 * ----------------______________________________________________
 *
 *   Copyright (C) 1995 Graham Wheeler, Francisco J. Ballesteros, 
 *                      Denis Froschauer
 *   Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach us by email to any of
 *    gram@aztec.co.za, nemo@ordago.uc3m.es, 100741.1151@compuserve.com
 */

#ident "@(#) LiS msg.c 2.1 4/4/99 11:34:24 "

/*
 * The memory allocation mechanism is based on that in SVR4.2.
 * That is, all memory is allocated dynamically, with freed
 * message headers being held on a free list. When kernel memory
 * is low some of these can be reclaimed by calling strgiveback.
 *
 * Message headers are 128 bytes in size. The extra space
 * is used as the data buffer for smallish messages. This
 * scheme means that in most cases, a call to allocb just
 * requires unlinking a message header from the free list
 * and initialising it.
 *
 * This scheme does add some complexity, however, with
 * regard to dupb/dupmsg. In this case the duplicate can
 * have pointers to a data buffer within some other message
 * header. Thus, if a message header is freed, we have to
 * check if its internal data buffer is still in use by
 * someone else, in which case we defer freeing the header;
 * on the other hand, if we are freeing the last reference
 * to some other data buffer in a message header, we have two
 * headers to free. All of this logic is nicely hidden in freeb()
 * (with a little bit of it leaking into pullupmsg).
 *
 * NB: This does rely on the fact that if a message block
 * has a data buffer of FASTBUF or less in size and no special
 * free function (i.e. it wasn't an esballoc), then that data
 * buffer lives internally within some (not necessarily the same)
 * message header, and was *not* allocated elsewhere.
 *
 * To put it another way, if you don't completely understand
 * the memory management scheme, don't fiddle with any of
 * the following code, and don't ever directly modify data
 * block elements like db_base, db_lim and db_size.
 */

/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */

#include <sys/stream.h>

/*  -------------------------------------------------------------------  */
/*			       Symbols & types                           */



/*  -------------------------------------------------------------------  */
/*				  Glob. Vars                             */

struct mdbblock  *lis_mdbfreelist = NULL; /* message block free list	 */
int lis_strmsgsz;	 		  /* maximum stream message size */
int lis_strcount ;			  /* # bytes allocated to msgs   */
long	 lis_max_msg_mem ;		  /* maximum to allocate */

/*  -------------------------------------------------------------------  */
/*			   Local functions & macros                      */


/* allochdr - allocate a message header. For internal use.
 *
 * Use DBLK_ALLOC to get the memory.  Since a header (mblk) also
 * contains a small data block it must be DMA-able memory.
 */
static struct mdbblock *
allochdr(char *file_name, int line_nr)
{
    struct mdbblock *rtn;
    int psw;
    SPLSTR(psw);
    rtn = lis_mdbfreelist;
    /* Try to get an mdbblock from the free list */
    if (rtn == NULL) /* free list empty; try allocate a new header */
    {
	SPLX(psw);			/* don't need protection for alloc */
	if (lis_max_msg_mem == 0 || lis_strcount < lis_max_msg_mem)
	    rtn = (struct mdbblock *)
		    DBLK_ALLOC(sizeof(struct mdbblock), file_name, line_nr, 0);
	SPLSTR(psw);			/* need to protect counters below */
    }
    else
    {
	/* paranoia */
	if (lis_mdbfreelist ==
	    (struct mdbblock *)lis_mdbfreelist->msgblk.m_mblock.b_next)
	{
	    rtn = (struct mdbblock *) NULL;
	}
	else
	{
	    /* remove header from free list */
	    lis_mdbfreelist = 
		(struct mdbblock *)lis_mdbfreelist->msgblk.m_mblock.b_next;

	    LisDownCount(FREEHDRS);
	    lis_mark_mem(rtn, file_name, line_nr) ;     /* under new ownership*/
	}
    }

    if (rtn)	/* update counts of header allocations */
    {
	LisUpCount(HEADERS);
	rtn->msgblk.m_mblock.b_next = NULL;
    }
    else	/* update counts of header allocation failures */
    {
	LisUpFailCount(HEADERS);
    }

    SPLX(psw);
    return rtn;

}/*allochdr*/

/*  -------------------------------------------------------------------  */

/* freehdr - return a header to the free list. For internal use.
 */

static void
freehdr(mblk_t *bp)
{
    int psw;
    SPLSTR(psw);

    LisDownCount(HEADERS);

    if (lis_strstats[FREEHDRS][CURRENT] >= MAX_MBLKS)
    {
	SPLX(psw);
	FREE(bp);
	return ;
    }

    bp->b_cont = NULL ;
    bp->b_next = (mblk_t *)lis_mdbfreelist;
    /*
     * If the data block ptr points to the data area of this same
     * mblk then leave it alone for port mortem debugging purposes.
     * If it points outside this area then clear these pointers
     * so that memory printers won't be looking at deallocated memory.
     */
    if (   bp->b_datap == NULL
	|| (bp->b_datap->db_base != (void*) ((struct mdbblock *)bp)->databuf)
       )
    {
	bp->b_datap= NULL ;			/* clear data block ptr */
	bp->b_rptr = NULL ;
	bp->b_wptr = NULL ;
    }

    lis_mdbfreelist = (struct mdbblock *)bp;
    lis_mark_mem(bp, "free", MEM_MSG) ;		     /* under new ownership */
    LisUpCount(FREEHDRS);
    SPLX(psw);
}/*freehdr*/

/*  -------------------------------------------------------------------  */
/* lis_msg_terminate - do the final shutdown of the msg memory subsystem.
 *
 * This will clean up any memory leaks.					*
 */
void
lis_terminate_msg(void)
{
    struct mdbblock *p;

    while (lis_mdbfreelist != NULL)
    {
	p = lis_mdbfreelist;
	lis_mdbfreelist = 
		(struct mdbblock *)lis_mdbfreelist->msgblk.m_mblock.b_next;
	FREE(p);
    }
}


/*  -------------------------------------------------------------------  */
/* allocdb - allocate a data buffer. For internal use.
 *
 * Use DBLK_ALLOC to allocate the memory.  STREAMS buffers must
 * be DMA-able.
 */
static char *
allocdb(int size, char *file_name, int line_nr, int flags)
{
    char *rtn;
    int   psw;

    if (lis_max_msg_mem != 0 && lis_strcount >= lis_max_msg_mem)
    {
	SPLSTR(psw);
	LisUpFailCount(DATABS);
	SPLX(psw);
	return(NULL) ;
    }

    ASSERT(size > FASTBUF);
    rtn = (char *)DBLK_ALLOC(size, file_name, line_nr, flags);
    SPLSTR(psw);
    if (rtn == NULL)
	LisUpFailCount(DATABS);
    else
    {
	lis_strcount += size;/* increment count of allocated memory */
	LisUpCount(DATABS);
	LisUpCounter(DBLKMEM, size);
    }

    SPLX(psw);
    return rtn;
}/*allocdb*/

/*  -------------------------------------------------------------------  */
/* initb: initialise an allocated msg block. For internal use only.
 */

static void
initb(struct mdbblock *blk, char *buff, int size,
      struct free_rtn *frtn)
{
    static struct mbinfo	mb_zero ;
    static frtn_t		frtn_zero ;

    /* m_func field is unused */
    blk->msgblk = mb_zero ;
    blk->msgblk.m_mblock.b_rptr = (unsigned char *)buff;
    blk->msgblk.m_mblock.b_wptr = (unsigned char *)buff;
    blk->msgblk.m_mblock.b_datap=  (struct datab *)&blk->datblk.d_dblock;
    
    if (frtn != NULL)
    {
	blk->datblk.d_dblock.db_rtn  = *frtn ;	/* save structure */
	blk->datblk.d_dblock.frtnp = &blk->datblk.d_dblock.db_rtn;
    }
    else
    {
	blk->datblk.d_dblock.db_rtn = frtn_zero ;
	blk->datblk.d_dblock.frtnp = NULL ;
    }

    blk->datblk.d_dblock.db_base = (unsigned char *)buff;
    if ((char *) buff != blk->databuf)	/* separate data block */
    {					/* of exact size */
	blk->datblk.d_dblock.db_lim = (unsigned char *)(buff + size);
	blk->datblk.d_dblock.db_size = size;
    }
    else				/* data block w/header */
    {
	blk->datblk.d_dblock.db_lim = (unsigned char *)(buff + FASTBUF);
	blk->datblk.d_dblock.db_size = FASTBUF;
    }

    blk->datblk.d_dblock.db_ref = (uchar)1;
    blk->datblk.d_dblock.db_type = (uchar)M_DATA;
    if (LIS_DEBUG_CLEAN_MSG)
	lis_bzero(blk->datblk.d_dblock.db_base, blk->datblk.d_dblock.db_size) ;
}/*initb*/

/*  -------------------------------------------------------------------  */
/*			Exported functions & macros                      */


/*  -------------------------------------------------------------------  */

/* allocb_physreq: allocate an M_DATA message block of the specified
 *	size and memory constraints. 
 *
 * NOTE:  This routine is not implemented correctly.  physreq_ptr is
 *        supposed to point to a structure that describes the memory
 *        attributes desired.  See SVR4 man pages for details.
 *
 *        To be completed later.
 */

struct msgb *
lis_allocb_physreq(int size, unsigned int priority, void *physreq_ptr,
			char *file_name, int line_nr)
{
    return(lis_allocb(size, priority, file_name, line_nr)) ;

}/*lis_allocb*/

/*  -------------------------------------------------------------------  */

/* allocb: allocate an M_DATA message block of the specified
 *	size. The priority is for compatibility only.
 */

struct msgb *
lis_allocb(int size, unsigned int priority, char *file_name, int line_nr)
{
    struct mdbblock *rtn = allochdr(file_name, line_nr);
    char *buff;
    (void)priority;
    if (rtn == NULL)
	return NULL;
    rtn->msgblk.m_mblock.b_datap = NULL ;     /* in case we need to free it */
    if (size > FASTBUF)
    {
	/* allocate data buffer */
	if ((buff = allocdb(size, file_name, line_nr, 0)) == NULL)
	{
	    /* failed; replace header in free list and return */
	    freehdr((mblk_t *)rtn);
	    return NULL;
	}
    }
    else buff = rtn->databuf;

    /* Set up message and data block headers */

    initb(rtn, buff, size, NULL);

    if ( LIS_DEBUG_ALLOC )
    {
	if (LIS_DEBUG_ADDRS)
	  printk("lis_allocb: mp=0x%lx dp=0x%lx(%s) from %s %d"
	         "rptr=0x%lx wptr=0x%lx size=%d\n",
		  (long) rtn,
		  (long) rtn->msgblk.m_mblock.b_datap,
		  (size > FASTBUF ? "extl" : "intl"),
		  file_name, line_nr,
		  (long) ((mblk_t *)rtn)->b_rptr,
		  (long) ((mblk_t *)rtn)->b_wptr,
		  size);
	else
	  printk("lis_allocb: from %s #%d size=%d\n", file_name, line_nr, size);
    }
    return (mblk_t *)rtn;
}/*lis_allocb*/

/*  -------------------------------------------------------------------  */
/* testb: see if an allocation can actually be done.
 */

int
lis_testb(int size, unsigned int priority)
{
#ifdef AVAIL
    int sz = 0;
    (void)priority;
    if (size > FASTBUF)
	sz += size;
    if (lis_mdbfreelist == NULL)
	sz += sizeof(struct mdbblock); /* gotta getta hedda */
    return (sz <= AVAIL());
#else
    /* hack - allocate and free */
    mblk_t *bp = allocb(size, priority);
    return bp ? (lis_freeb(bp), 1) : 0;
#endif
}/*lis_testb*/	

/*  -------------------------------------------------------------------  */
/*
 * esballoc: allocate a message block, using a user-provided data buffer
 */

mblk_t *
lis_esballoc(unsigned char *base, int size, int priority,
	     frtn_t *freeinfo, char *file_name, int line_nr)
{
    struct mdbblock *rtn = NULL;
    (void)priority;
    if (base && freeinfo && (rtn  = allochdr(file_name, line_nr)) != NULL)
    	initb(rtn, base, size, freeinfo);
    return (mblk_t *)rtn;
}/*lis_esballoc*/

/*  -------------------------------------------------------------------  */

/* lis_freedb - Free the data block associated with a message and
 *              possibly the message header as well.
 */
void
lis_freedb(mblk_t *bp, int free_hdr)
{
    int psw;

    if (bp)
    {
	struct datab *dp = bp->b_datap;

    	if (dp)
	{
	    if (dp -> db_ref == 0)
	    {
		printk ("lis_freeb: mp 0x%lx ref count 0\n", (long) bp);
		return;
	    }

	    if (--dp->db_ref == 0)
            {
	    	if (dp->frtnp)
		{
		    struct mdbblock	*blk = (struct mdbblock *) bp ;

		    if ( LIS_DEBUG_FREEMSG && LIS_DEBUG_ADDRS )
		       printk("lis_freeb: (user data blk) dp=0x%lx(extl)"
		              " dp->db_base=0x%lx dp->db_lim=0x%lx"
			      " dp->db_size=%d\n",
			       (long) dp, (long) dp->db_base,
			       (long) dp->db_lim, dp->db_size);

	            (*(dp->frtnp->free_func))(dp->frtnp->free_arg);
		    bp->b_datap = (struct datab *)(&blk->datblk.d_dblock);
		}
	    	else
		if (dp->db_size > FASTBUF)
	    	{				/* separate data buffer */
		    if ( LIS_DEBUG_FREEMSG && LIS_DEBUG_ADDRS  )
			printk("lis_freeb: (data blk) dp=0x%lx(extl)"
		               " dp->db_base=0x%lx dp->db_lim=0x%lx"
			       " dp->db_size=%d\n",
			       (long) dp, (long) dp->db_base,
			       (long) dp->db_lim, dp->db_size);

		    SPLSTR(psw);
	            lis_strcount -= dp->db_size;
	            LisDownCount(DATABS);
		    LisDownCounter(DBLKMEM, dp->db_size);
		    SPLX(psw);

	            FREE(dp->db_base);		/* free data area */
	        }

		/*
		 * The data block holding onto this area may not be
		 * in the 'bp' buffer.  If not, then free the msgblk
		 * that contains the data block since its ref count
		 * is now zero.
		 *
		 * We get to the msgblk by starting at the embedded
		 * data block and backing up over the struct mbinfo
		 * that precedes the data block.
		 */
		if (dp != (dblk_t *) &((struct mdbblock *)bp)->datblk)
		{
		    mblk_t		*dpbp ;
		    struct mdbblock	*blk = (struct mdbblock *) bp ;

		    dpbp = (mblk_t *) ( ((struct mbinfo *)dp) - 1 ) ;
		    if ( LIS_DEBUG_FREEMSG && LIS_DEBUG_ADDRS )
			printk("lis_freeb: (mblk[dblk]) mp=0x%lx dp=0x%lx\n",
					(long) dpbp, (long) dp);
		    freehdr(dpbp);
		    bp->b_datap = (struct datab *)(&blk->datblk.d_dblock);
		}

		/*
		 * Either the data block is internal to the mblk and will
		 * be freed along with the header, or the data area was
		 * external and has already been freed by the code just
		 * above.
		 */

		if (free_hdr)
		{
		    if ( LIS_DEBUG_FREEMSG && LIS_DEBUG_ADDRS  )
			printk("lis_freeb: (mblk) bp=0x%lx dp=0x%lx\n",
			(long) bp, (long) bp->b_datap);

		    freehdr(bp); 	/* free the header */
		}
	    }
	    /* 
	     * If the ref count is non-zero then check to see of the
	     * data block is in the same message as 'bp'.  If not then
	     * it is safe to free 'bp' but not the msg block in which
	     * the data block resides.  That will be done later when
	     * the ref count goes to zero.
	     */
	    else
	    if (   free_hdr
		&& (dp != (dblk_t *) &((struct mdbblock *)bp)->datblk)
	       )
	    {
		if ( LIS_DEBUG_FREEMSG && LIS_DEBUG_ADDRS  )
		    printk("lis_freeb: (mblk) bp=0x%lx dp=0x%lx\n",
			    (long) bp, (long) bp->b_datap);

        	freehdr(bp);
	    }
        }
        else
	if (free_hdr)
	{
	    if ( LIS_DEBUG_FREEMSG && LIS_DEBUG_ADDRS  )
		printk("lis_freeb: (mblk/no-dblk) bp=0x%lx dp=NULL\n",
			(long) bp);

	    freehdr(bp);	 /* no data block; just free the header */
	}
    }

}/*lis_freedb*/

/*  -------------------------------------------------------------------  */

/* freeb - Free data buffer and place message block on free list. Don't
 *      follow the continuation pointer.
 */
void
lis_freeb(mblk_t *bp)
{
    lis_freedb(bp, 1) ;			/* free both hdr and data blk */

}/*lis_freeb*/

/*  -------------------------------------------------------------------  */
/* freemsg - free a whole message
 */
void
lis_freemsg(mblk_t *mp)
{
    while (mp)
    {
	mblk_t *now = mp;
	mp = mp->b_cont;
	lis_freedb(now, 1);
    }
}/*lis_freemsg*/

/*  -------------------------------------------------------------------  */
/* extract values from a char pointer and advance the pointer */
int lis_getint(unsigned char **p)
{
    int result = *((int*)(*p));
    *p += sizeof(int);
    return(result);
}

/*  -------------------------------------------------------------------  */
void lis_putbyte(unsigned char **p, unsigned char byte)
{
    **p = byte;
    ++*p;
}

/*----------------------------------------------------------------------
# Local Variables:      ***
# change-log-default-name: "~/src/prj/streams/src/NOTES" ***
# End: ***
  ----------------------------------------------------------------------*/

