/************************************************************************
*                  Portable Machine Dependent Code                      *
*************************************************************************
*									*
* This file contains routines for machine (or O/S) dependent		*
* functions for the portable version of the streams code.		*
*									*
* This file contains some routines that will port to any environment	*
* and includes a second .c file which will contains routines that are	*
* unique to a particular environment.					*
*									*
* Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>		*
*									*
************************************************************************/
/*
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 */

#ident "@(#) LiS port-mdep.c 2.1 4/4/99 11:34:24 "

#include <sys/stream.h>
#include <sys/LiS/errmsg.h>


/************************************************************************
*                         Global Variables                              *
************************************************************************/

unsigned long		port_clock ;	/* milli-second clock */

/************************************************************************
*                           port_init                                   *
*************************************************************************
*									*
* This routine is called from outside to initialize the STREAMS system.	*
*									*
************************************************************************/
void	port_init(void)
{
    extern void		lis_init_head(void) ;

    lis_init_head() ;

} /* port_init */

/************************************************************************
*                         port_copyin_str                               *
*************************************************************************
*									*
* Portable routine to copy a string in from user space to kernel.	*
*									*
* ustr		pointer to string in user space.			*
* kstr		pointer to a pointer that will point to kernel memory	*
*		where the string is copied.  We allocate this memory	*
*		and the caller must free it using the FREE macro.	*
* max		Max length of the string.				*
*									*
* Return:	< 0 for error, caller does not free memory.		*
*		0   for success, caller must free memory.		*
*									*
************************************************************************/
int 
port_copyin_str(struct file *f, const char *ustr, char **kstr, int max)
{
	int    i = max;
	int    error;
	char   c;
	char  *tmp;
	char  *mem ;

	if (max <= 0)
	    return(-ENOMEM);

	error = -EFAULT;
	c = (char) port_get_fs_byte(f, ustr++);
	if (!c)
	    return(-EINVAL);

	if ((mem = ALLOCF(max,"copyin-buf ")) == NULL)
	    return(-ENOMEM);

	*kstr = tmp = mem ;
	while (--i)
	{
	    *(tmp++) = c;
	    c = (char) port_get_fs_byte(f, ustr++);
	    if (!c)
	    {
		*tmp = '\0';
		return(0);
	    }
	}

	FREE(mem);
	return(error);

} /* port_copyin_str */

/************************************************************************
*                            lis_errmsg                                 *
*************************************************************************
*									*
* Return a pointer to a string that will serve as a prefix for the	*
* standard lis error message.						*
*									*
************************************************************************/
char  *lis_errmsg( int lvl )
{
    switch(lvl)
    {
    case LIS_PANIC:
	return("PANIC");
    case LIS_ERROR:
	return("ERROR");
    case LIS_WARN:
	return("WARNING");
    case LIS_DEBUG:
	return("DEBUG");
    default:
	return("unknown severity");
    }

} /*lis_errmsg*/

/************************************************************************
*                         Timer Mechanism                               *
*************************************************************************
*									*
* This code is borrowed from the Linux kernel and reworked so that	*
* it maintains time deltas rather than absolute time.  See the comments	*
* in the individual routines below.					*
*									*
* Author:	David Grothe <dave@gcom.com>				*
*									*
************************************************************************/

static struct timer_list timer_head =
			 { &timer_head, &timer_head, 0, 0, NULL };

/*
 * Add a timer list entry to the list of entries, this effectively
 * starts the timer.  The list is kept in ascending order of number
 * of ticks until the timer expires.  As you go down the list the
 * 'tdelta' fields are cumulative.  Each one specifies the additional
 * amount of time after the previous entry would expire.
 */
void port_add_timer(struct timer_list * timer)
{
    int			 flags;
    struct timer_list	*p;

    if (timer->next != NULL || timer->prev != NULL)
    {
	printk("add_timer(%lx) called with non-zero list\n", timer) ;
	return;
    }

    p = &timer_head;
    SPLSTR(flags);
    do
    {
	timer->tdelta -= p->tdelta ;	/* reduce ticks difference */
	p = p->next;

    } while (p != &timer_head && timer->tdelta > p->tdelta);

    if (p != &timer_head)		/* real entry */
	p->tdelta -= timer->tdelta;	/* less delta to get to the next one */

    timer->next = p;			/* put timer before 'p' */
    timer->prev = p->prev;
    p->prev = timer;
    timer->prev->next = timer;

    SPLX(flags);

} /* port_add_timer */

/*
 * Find the timer-list entry and remove it from the linked list
 */
int port_del_timer(struct timer_list * timer)
{
    int			 flags;
    struct timer_list	*p;

    p = &timer_head;
    SPLSTR(flags);
    while ((p = p->next) != &timer_head)
    {
	if (p == timer)
	{
	    if (p->next != &timer_head)		/* real timer entry follows */
	    	p->next->tdelta += timer->tdelta ;	/* add back our time */

	    timer->next->prev = timer->prev;	/* delink our entry */
	    timer->prev->next = timer->next;
	    timer->next = timer->prev = NULL;

	    /*
	     * Clear any residual ticks saved up by the
	     * announce time routine if the timer list is
	     * now empty.
	     */
	    if (timer_head.next == &timer_head)
		port_announce_time(0) ;

	    SPLX(flags);
	    return 1;
	}
    }

    if (timer->next || timer->prev)
	    printk("del_timer(%lx) called from with timer not found in table\n",
		    timer) ;

    SPLX(flags);
    return 0;

} /* port_del_timer */

/*
 * The porting environment must call this routine periodically to
 * input time into the timeout mechanism.  This routine can be called
 * from interrupt level or from background.  The entire routine runs
 * with interrupts disabled.
 *
 * Calls to user time out routines will be made from here with the
 * processor status restored to that of caller.  If you call this
 * routine from interrupt level then all the timer callout routines
 * will be called with interrupts disabled.  If you call it from
 * background then they will be run with interrupts enabled.
 */
void port_announce_time(long new_milli_secs)
{
    static unsigned long milli_secs ;	/* time interval saved up */
    static int		 entered ;
    int			 flags ;
    struct timer_list	*timer;

    SPLSTR(flags);			/* protect linked list loop */

    milli_secs += new_milli_secs ;	/* add to total nr of milli-secs */
    port_clock += new_milli_secs ;	/* keep global elapsed time */

    if (timer_head.next == &timer_head)	/* list empty */
    {
	milli_secs = 0 ;		/* do not accumulate time */
	SPLX(flags);			/* re-enable intrs */
	return ;			/* nested call */
    }

    if (entered)			/* nested entry */
    {
	SPLX(flags);			/* re-enable intrs */
	return ;			/* nested call */
    }

    entered = 1 ;			/* prevent nesting */

    /*
     * Because the table is ordered by time deltas, it is always the
     * ones at the head of the list that need to be sprung.  Whenever
     * we spring one we decrement that time's tdelta from our number
     * of milliseconds that have passed so we can meaningfully compare
     * against the next entry.
     */
    while (   (timer = timer_head.next) != &timer_head	/* list non-empty */
	   && timer->tdelta <= milli_secs		/* 1st elt expired */
	  )
    {
	void		       (*fn)(unsigned long) = timer->function;
	unsigned long     	 data = timer->data;

	milli_secs 	     -= timer->tdelta ;	/* less time to nxt ntry */

	timer->next->prev     = timer->prev;	/* delink timer entry */
	timer->prev->next     = timer->next;
	timer->next	      =			/* clear links */
	timer->prev	      = NULL;

	SPLX(flags);
	if (fn != NULL) fn(data);		/* be careful of NULL ptr */
	SPLSTR(flags);
    }

    entered = 0 ;				/* allow calls again */

    SPLX(flags);
	
} /* port_announce_time */

/*
 * This routine is called from dki.c to start a timer. 'ticks' is
 * in units of HZ (typically 1/100th of a second, or 10ms).
 */
void
lis_tmout(struct timer_list *tl, void (*fn)(ulong), long arg, long ticks)
{
    tl->next	= NULL ;
    tl->prev	= NULL ;
    tl->function= fn;
    tl->data	= arg;
    tl->tdelta	= ticks * (1000/HZ) ;		/* convert to milli-secs */
    port_add_timer(tl);
}

/*
 * This routine is called from dki.c to stop a timer.
 */
void
lis_untmout(struct timer_list *tl)
{
    port_del_timer(tl);
}

/************************************************************************
*                            port_time_till                             *
*************************************************************************
*									*
* Given a target time in terms of elapsed milli-seconds, in other words,*
* the same units as the global port_clock, return the number of		*
* milli-seconds that it will take to reach to target time.		*
*									*
* The return is negative if the target time is in the past, zero	*
* if it is the same as the current target time, and positive if		*
* target time is in the future.						*
*									*
************************************************************************/
long	port_time_till(long target_time)
{
    return( target_time - port_clock ) ;

} /* port_time_till */

/************************************************************************
*                           port_target_time                            *
*************************************************************************
*									*
* Convert the milli_sec interval to an absolute target time expressed	*
* in the same manner as the global port_clock.				*
*									*
************************************************************************/
long	port_target_time(long milli_sec)
{
    return( port_clock + milli_sec ) ;

} /* port_target_time */
