/*                               -*- Mode: C -*- 
 * queue.c --- streams statistics
 * Author          : Graham Wheeler, Francisco J. Ballesteros
 * Created On      : Tue May 31 22:25:19 1994
 * Last Modified By: David Grothe
 * RCS Id          : $Id: queue.c,v 1.9 1996/01/24 19:01:26 dave Exp $
 * Purpose         : provide some queue for LiS
 * ----------------______________________________________________
 *
 *   Copyright (C) 1995  Graham Wheeler, Francisco J. Ballesteros
 *   Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach us by email to 
 *    gram@aztec.co.za, nemo@ordago.uc3m.es
 *    dave@gcom.com
 */

#ident "@(#) LiS queue.c 2.4 6/17/99 14:47:30 "



/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */

#include <sys/stream.h>

/*  -------------------------------------------------------------------  */
/*			   Local functions & macros                      */


/*  -------------------------------------------------------------------  */
/* find_qband
 *
 * Given a queue and a class number, return a pointer to the
 * qband structure for that band.
 *
 * Return NULL if we had to allocate a band structure and we
 * could not get the memory.
 *
 * This routine assumes that SPLSTR has been done before calling it.
 */
static INLINE struct qband *
find_qband(queue_t *q, int msg_class)
{
    struct qband	*qp ;
    uchar		 msgclass = (uchar) msg_class ;

    if (q == NULL || msgclass > LIS_MAX_BAND) return(NULL) ;

    qp = q->q_bandp;
    while (qp != NULL)
    {
	if ( qp->qb_first != NULL && qp->qb_first->b_band == msgclass )
	    break;			/* found the band structure */

	qp = qp->qb_next;
    }

    if (qp == NULL)			/* could not find the band struct */
    {					/* find an unused one */
	qp = q->q_bandp;
	while (qp != NULL)
	{
	    if ( qp->qb_first == NULL )
		break;			/* unused band structure */

	    qp = qp->qb_next;
	}
    }

    if (qp == NULL)			/* no such band */
    {					/* and no unused structures */
	qp = (qband_t*) ALLOCF(sizeof(struct qband),"qband ");
	if (qp == NULL) return(NULL) ;	/* could not allocate */

	qp->qb_next = q->q_bandp;
	q->q_bandp = qp;
	qp->qb_first = qp->qb_last = NULL;
    }

    if (qp->qb_first == NULL)
    {					/* new or resused qband structure */
	qp->qb_count = 0;
	qp->qb_flag  = 0 ;
	qp->qb_hiwat = q->q_hiwat;
	qp->qb_lowat = q->q_lowat;
    }

    return(qp) ;			/* rtn ptr to band structure */

} /* find_qband */

/*  -------------------------------------------------------------------  */
/* updatequeue - update queue info after a message is added
 *	Assumes that SPLSTR has been done!
 *
 * This routine assumes that SPLSTR has been done before calling it.
 */
static INLINE void
updatequeue(queue_t *q, mblk_t *mp, int from_insq)
{
    uchar	 msg_class = mp->b_band;
    uchar	 msg_type  = mp->b_datap->db_type ;

    if (msg_type >= QPCTL || mp->b_band == 0)
    {
        q->q_count += lis_msgsize(mp);
        if (q->q_count > q->q_hiwat)
	    q->q_flag |= QFULL | QWASFULL ;
    }
    else
    {						/* qband considerations */
	struct qband *qp = find_qband(q, msg_class) ; /* get qband struct */

	if (qp != NULL)
	{
	    qp->qb_count += lis_msgsize(mp);	/* qband stats */
	    if (qp->qb_count > qp->qb_hiwat)
		qp->qb_flag |= QB_FULL | QB_WASFULL ;

	    if (qp->qb_first == NULL)
		qp->qb_first = mp;		/* this is 1st msg in band */

	    if (qp->qb_last == NULL)		/* new qband */
		qp->qb_last = mp;
	}
    }

    if (from_insq && !lis_canenable(q))		/* noenable in effect */
	return ;				/* insq can't cause sched */

    if (msg_type >= QPCTL ||
       (lis_canenable(q) && (q->q_flag & QWANTR) != 0))
         lis_qenable(q);

}/*updatequeue*/

/*  -------------------------------------------------------------------  */
/* find_q_spot
 *
 * Find "the correct spot" in the queue for the message.
 *
 * Find the element that we will link this message in front of. 
 *
 * find_q_spot_head finds the first message of the proper class.
 * find_q_spot_tail finds the message beyond the last message of
 * the given class.
 *
 * Return NULL if there is no message that we can link in front of.
 * This will be the return for an empty queue, or for a message that
 * needs to be placed on the end of the queue.
 *
 * This routine assumes that SPLSTR has been done before calling it.
 */
static INLINE mblk_t *
find_q_spot_head(queue_t *q, mblk_t *mp)
{
    uchar	 msg_class ;
    uchar	 mp_type ;
    mblk_t	*p_cont ;
    mblk_t	*firstp ;

    if (q == NULL || (firstp = q->q_first) == NULL) return(NULL) ;

    msg_class = mp->b_band;
    if ((mp_type = mp->b_datap->db_type) >= QPCTL)
	mp->b_band = 0 ;			/* AT&T STRMS pgrmrs guide */

    if (mp_type >= QPCTL)			/* msg is high priority */
	return(firstp) ;			/* goes as the 1st msg */

    if (   firstp->b_datap->db_type < QPCTL	/* 1st is not high prior */
	&& msg_class >= firstp->b_band		/* greater class than 1st msg */
       )
	    return(firstp) ;			/* goes as 1st msg */

    /* find the first non-high-priority message in the queue */
    for (p_cont = firstp;
	 (   p_cont != NULL
	  && p_cont->b_datap->db_type >= QPCTL   /* skip high prior msgs */
	 );
	 p_cont = p_cont->b_next
	) ;

    /* find the first message after that with class <= our msg */
    for (;
	 (   (p_cont != NULL)
	  && (p_cont->b_band > msg_class)	/* skip higher band msgs */
	 );
	 p_cont = p_cont->b_next
	) ;

    return(p_cont) ;			/* insert before here */

} /* find_q_spot_head */

static INLINE mblk_t *
find_q_spot_tail(queue_t *q, mblk_t *mp)
{
    uchar	 msg_class ;
    uchar	 mp_type ;
    mblk_t	*p_cont ;
    mblk_t	*lastp ;
    mblk_t	*firstp ;

    if (q == NULL || (firstp = q->q_first) == NULL) return(NULL) ;

    msg_class = mp->b_band;
    if ((mp_type = mp->b_datap->db_type) >= QPCTL)
	mp->b_band = 0 ;			/* AT&T STRMS pgrmrs guide */

        /* find a spot at the end of a message class */
    lastp = q->q_last ;
    if (lastp->b_datap->db_type >= QPCTL)	/* q full of hi-prior msgs */
	    return(NULL) ;			/* it goes on the end */

    if (   mp_type < QPCTL			/* non-priority msg */
        && msg_class <= lastp->b_band		/* lower than last msg */
       )
	    return(NULL) ;			/* it goes on the end */

    /* have to find a spot in the middle of the queue */
    /* first, skip over all of the high-priority msgs */

    for (p_cont = firstp;
	 (   p_cont != NULL
	  && p_cont->b_datap->db_type >= QPCTL   /* skip high prior msgs */
	 );
	 p_cont = p_cont->b_next
	) ;

    /* if the message is a high priority message then we are done */

    if (mp_type < QPCTL)			/* non-priority msg */
    {
    	for (;					/* find end of mp's band */
    	     (   (p_cont != NULL)
	      && (p_cont->b_band >= msg_class)	/* skip higher band msgs */
	     );
	     p_cont = p_cont->b_next
	    ) ;
    }

    return(p_cont) ;				/* insert before here */

} /* find_q_spot_tail */


/*  -------------------------------------------------------------------  */
/* ins_before
 *
 * This routine will insert the given message in front of the
 * message whose ptr is passed in.
 *
 * This routine assumes that SPLSTR has been done before calling it.
 */
static INLINE void
ins_before(queue_t *q, mblk_t *before_msg, mblk_t *mp)
{
    LisUpCount(MSGSQD) ;			/* one more msg queued */
    if (q->q_first == NULL)			/* empty queue */
    {
    	mp->b_prev = mp->b_next = NULL;		/* quick & easy */
    	q->q_first = q->q_last = mp;
    	return ;				/* done */
    }

    if (before_msg == NULL)			/* add to end of queue */
    {
	q->q_last->b_next	= mp;
	mp->b_prev		= q->q_last;
	q->q_last		= mp;
	mp->b_next		= NULL;
	return ;
    }

    /* insert in front of 'before_msg' */

    mp->b_next			   = before_msg;
    mp->b_prev			   = before_msg->b_prev;

    if (before_msg->b_prev != NULL)
	before_msg->b_prev->b_next = mp;
    else
	q->q_first		   = mp;

    before_msg->b_prev		   = mp;

} /* ins_before */

/*  -------------------------------------------------------------------  */
/* rmv_msg
 *
 * Remove a message from a queue.  The message is known to reside in
 * the queue.
 *
 * This routine assumes that SPLSTR has been done before calling it.
 */
static INLINE mblk_t *
rmv_msg(queue_t *q, mblk_t *mp)
{
    if (mp == NULL)			/* no message */
    {					/* empty queue */
	q->q_flag |= QWANTR | QBACK;	/* note that someone wants to read */
	lis_backenable(q);
	return(mp) ;
    }

    q->q_flag &= ~QWANTR;		/* no implicit qenable now */

    if (mp->b_datap->db_type < QPCTL && mp->b_band > 0)
    {
	struct qband *qp = find_qband(q, mp->b_band) ; /* get qband struct */

	if (qp != NULL)
	{
	    qp->qb_count -= lis_msgsize(mp);	/* qband stats */
	    if (qp->qb_count <= qp->qb_hiwat)
		qp->qb_flag &= ~QB_FULL;

	    if (   (qp->qb_flag & QB_WASFULL)	/* q was full */
	        && qp->qb_count <= qp->qb_lowat	/* now below low water?	*/
	       )
	    {
		/* Does anyone know what it means to back-enable a band? */
		/* I will assume that we do the same type of processing  */
		/* as for the whole queue.				 */
		/* One bug is that nobody ever clears the QB_BACK flag!	 */

		qp->qb_flag |= QB_BACK;		/* must back enable	*/
		qp->qb_flag &= ~QB_WASFULL;	/* no longer full	*/
		if (qp->qb_flag & QB_WANTW)
		{				/* xfer flag to main queue */
		    qp->qb_flag &= ~QB_WANTW;
		    q->q_flag   |= QWANTW;
		}

		q->q_flag |= QBACK;
	    }

	    /* mp may be in the middle or end of a band if this was	*/
	    /* done via rmvq().						*/

	    if (qp->qb_first == mp)		/* rmving 1st msg of band */
	    {
		if (mp->b_next && mp->b_next->b_band == mp->b_band)
		    qp->qb_first = mp->b_next ;	/* nxt msg in same band */
		else
		{
		    qp->qb_first = NULL ;	/* band now empty */
		    qp->qb_last  = NULL;	/* band struct available */
		}
	    }

	    if (qp->qb_last == mp)		/* rmving last msg of band */
	    {
		if (mp->b_prev && mp->b_prev->b_band == mp->b_band)
		    qp->qb_last = mp->b_prev ;	/* nxt msg in same band */
		else
		{
		    qp->qb_first = NULL ;	/* band now empty */
		    qp->qb_last  = NULL;	/* band struct available */
		}
	    }
	}
    }
    else
    {
	q->q_count -= lis_msgsize(mp);
	if (q->q_count < q->q_hiwat)
	    q->q_flag &= ~QFULL;

	if (   (q->q_flag & QWASFULL)	/* queue was full		*/
	    && q->q_count <= q->q_lowat	/* below low water?		*/
	   )
	{
	    q->q_flag |= QBACK;		/* must back enable		*/
	    q->q_flag &= ~QWASFULL;	/* no longer full		*/
	}
    }

    /* perform q remove here */

    if (mp->b_prev != NULL)
    	mp->b_prev->b_next = mp->b_next;
    else
    	q->q_first = mp->b_next;

    if (mp->b_next != NULL)
    	mp->b_next->b_prev = mp->b_prev;
    else
    	q->q_last = mp->b_prev;

    mp->b_prev = mp->b_next = NULL;

    lis_backenable(q);
    LisDownCount(MSGSQD) ;		/* one fewer msg queued */

    return(mp) ;			/* return our argument */

} /* rmv_msg */


/*  -------------------------------------------------------------------  */
/*				  Glob. Vars                             */


/*  -------------------------------------------------------------------  */
/*			Exported functions & macros                      */

/*
 * lis_check_q_magic
 *
 * Return 1 if magic number in queue is OK.  Return 0 if not.
 */
int	lis_check_q_magic(queue_t *q, char *file, int line)
{
    if (q->q_magic != Q_MAGIC)
    {
	printk("%s #%u: Queue magic number is 0x%lx, should be 0x%lx\n",
		file, line, q->q_magic, Q_MAGIC) ;

	return(0) ;
    }

    return(1) ;

} /* lis_check_q_magic */

/* backq - Return the queue which feeds this one.
 *
 * If q is a read queue then return the ptr to the downstream queue.
 * If q is a write queue then return the ptr to the upstream queue.
 */
queue_t *
lis_backq(queue_t *q)
{
    queue_t	*other_q ;

    if (q == NULL || (other_q = LIS_OTHERQ(q)) == NULL)
    	return NULL;

    if (other_q->q_next == NULL)	/* is there a successor queue */
	return(NULL) ;			/* no */

    return(LIS_OTHERQ(other_q->q_next)) ;  /* the queue that points to 'q' */

}/*lis_backq*/

/*  -------------------------------------------------------------------  */

/* lis_backenable - enable back queue if QWANTW is set, i.e.
 *	if a back queue is full.
 *
 *	If the queue is marked as closing then do not stop
 *	at the first queue with a service procedure, keep
 *	going as long as the queues are empty.  The idea is
 *	to eventually enable the service procedure for the
 *	queue head.
 */
void 
lis_backenable(queue_t *q)
{
    if (   (   (q->q_flag & (QWANTW|QBACK))==(QWANTW|QBACK)
            || (q->q_flag & QCLOSING)
           )
	&& q->q_count<=q->q_lowat
       )
    {
	/* Someone wants to write and there is space. Find a previous
	   queue to back-enable and enable it */
    	q->q_flag &= ~(QWANTW|QBACK);
        while((q = lis_backq(q)) != NULL)
	{
	    if (   LIS_CHECK_Q_MAGIC(q)
		&& q->q_qinfo != NULL
		&& q->q_qinfo->qi_srvp != NULL)
	    {
	     	lis_qenable(q);
	     	if (!F_ISSET(q->q_flag,QCLOSING) || q->q_count != 0)
		    break;			/* quit searching */
	    }
	}
    }
}/*lis_backenable*/

/*  -------------------------------------------------------------------  */
/* lis_getq - get message from head of queue
 */

mblk_t *
lis_getq(queue_t *q)
{
    mblk_t *rtn;
    int     psw;

    if (q == NULL)
	return NULL;			/* sanity check	*/

    if (!LIS_CHECK_Q_MAGIC(q)) return(NULL) ;

    SPLSTR(psw);
    rtn = rmv_msg(q, q->q_first) ;
    SPLX(psw);
    return rtn;

}/*lis_getq*/

/*  -------------------------------------------------------------------  */
/* putq- put a message into a queue
 */
int
lis_putq(queue_t *q, mblk_t *mp)
{
    int psw;
    if (mp == NULL)
	return 0;

    if (   q == NULL
	|| !LIS_CHECK_Q_MAGIC(q)
       )
    {
    	lis_freemsg(mp);
    	return 0;
    }

    SPLSTR(psw);
    ins_before(q, find_q_spot_tail(q, mp), mp) ;
    updatequeue(q, mp, 0);		/* Update queue counts and flags */
    SPLX(psw);
    lis_runqueues();				/* schedule queues to run */
    return 1;					/* success */

}/*lis_putq*/


/*  -------------------------------------------------------------------  */
/* putbq - return a message to a queue
 */
int
lis_putbq(queue_t *q, mblk_t *mp)
{
    int psw;

    if (mp == NULL) return 0;

    if (   q == NULL
	|| !LIS_CHECK_Q_MAGIC(q)
       )
    {
    	lis_freemsg(mp);
    	return 0;
    }

    SPLSTR(psw);
    ins_before(q, find_q_spot_head(q, mp), mp) ;
    updatequeue(q, mp, 0);		/* Update queue counts and flags */
    SPLX(psw);
    return 1;					/* success */

}/*lis_putbq*/

/*  -------------------------------------------------------------------  */
/*insq - insert mp before emp. If emp is NULL, insert at end
 *	of queue. If the insertion is out-of-order, the insert fails.
 *	Returns 1 on success; 0 otherwise.
 */
int
lis_insq(queue_t *q, mblk_t *emp, mblk_t *mp)
{
    int psw;

    if (mp == NULL) return 0;

    if (   q == NULL
	|| !LIS_CHECK_Q_MAGIC(q)
       )
    {
    	lis_freemsg(mp);
    	return 0;
    }

    SPLSTR(psw);

    /* check to see if this insertion is proper */

    if (   mp->b_datap->db_type < QPCTL		/* non-high priority msg */
        && emp != NULL				/* have emp */
        && (   emp->b_datap->db_type >= QPCTL	/* emp is high priority */
            || mp->b_band < emp->b_band		/* too high a class */
            || (   emp->b_prev != NULL
                && mp->b_band > emp->b_prev->b_band
               )
           )
       )
    {
	LisUpFailCount(MSGSQD) ;		/* error queueing msg */
	SPLX(psw);
	return(0) ;				/* failed */
    }

    ins_before(q, emp, mp) ;
    updatequeue(q, mp, 1);
    SPLX(psw);
    return 1;					/* success */

}/*lis_insq*/

/*  -------------------------------------------------------------------  */
/* rmvq - remove a message from a queue. If the message
 *		does not exist, panic.
 */
void
lis_rmvq(queue_t *q, mblk_t *mp)
{
    mblk_t *bp;
    int psw;
    if (mp == NULL) return;
    if (q == NULL)
    {
	printk("Null queue in rmvq");
	return ;
    }

    if (!LIS_CHECK_Q_MAGIC(q)) return ;

    SPLSTR(psw);
    /* check if mesage is in queue */
    bp = q->q_first;
    while (bp && bp != mp)
	bp = bp->b_next;
    if (bp == NULL)
    {
	SPLX(psw);
	printk("rmvq: message not in queue");
	return ;
    }

    /* Remove message from queue */
    rmv_msg(q, bp) ;
    SPLX(psw);

}/*lis_rmvq*/

/*  -------------------------------------------------------------------  */

/* lis_qenable - schedule a queue for service
 */
void 
lis_qenable(queue_t *q)
{
    int psw;

    if (!LIS_CHECK_Q_MAGIC(q)) return ;

    SPLSTR(psw);
    /* is there a service procedure? if not, just return */
    if (q == NULL || q->q_qinfo == NULL || q->q_qinfo->qi_srvp == NULL)
	goto done;
    /* if already enabled, just return */
    if ((q->q_flag & QENAB) != 0)
	goto done;
    /* enable the queue for servicing */
    q->q_flag |= QENAB;
    /* link into tail of list of scheduled queues (qtail) */
    q->q_link = NULL;
    if (lis_qhead == NULL)
	lis_qhead = q;
    else
    	lis_qtail->q_link = q;
    lis_qtail = q;		
    LisUpCount(QSCHEDS) ;
    lis_setqsched();
done:
    SPLX(psw);
}/*lis_qenable*/

/*  -------------------------------------------------------------------  */

/* lis_setq - Set queue variables
 */
void
lis_setq(queue_t * q, struct qinit *rinit, struct qinit *winit)
{
    if (rinit == NULL || winit == NULL)
    {
	printk("lis_setq: NULL pointer(s) to qinit structures\n") ;
	return ;
    }

    if (q)
    {
	int psw;

	if (!LIS_CHECK_Q_MAGIC(q)) return ;

	SPLSTR(psw);

        q->q_qinfo  = rinit;
        q->q_lowat  = rinit->qi_minfo->mi_lowat;
        q->q_hiwat  = rinit->qi_minfo->mi_hiwat;
        q->q_minpsz = rinit->qi_minfo->mi_minpsz; 
        q->q_maxpsz = rinit->qi_minfo->mi_maxpsz; 

        q = LIS_WR(q);
        q->q_qinfo  = winit;
        q->q_lowat  = winit->qi_minfo->mi_lowat;
        q->q_hiwat  = winit->qi_minfo->mi_hiwat;
        q->q_minpsz = winit->qi_minfo->mi_minpsz; 
        q->q_maxpsz = winit->qi_minfo->mi_maxpsz; 

	SPLX(psw);
    }
}/*lis_setq*/

/*  -------------------------------------------------------------------  */
/* lis_flushband - flush messages in a specified priority band.
 *	flag can be FLUSHDATA (flush only M_DATA, M_DELAY,
 *	M_PROTO, M_PCPROTO) or FLUSHALL.
 */
void
lis_flushband(queue_t *q, unsigned char band, int flag)
{
    int psw;

    if (q == NULL) return;

    if (!LIS_CHECK_Q_MAGIC(q)) return ;

    if (band > 0)
    {
	struct qband *qp ;
	mblk_t	     *mp ;
	mblk_t	     *p_next ;

	SPLSTR(psw);
	qp = find_qband(q, band) ;			/* get qband struct */
	for (mp = qp->qb_first; mp != NULL; )
	{
	    if (mp->b_band != band)
	    {
		printk("flushband: flushing band %d: mp->b_band=%d\n",
			band, mp->b_band) ;
		break ;			/* quit */
	    }

	    p_next = mp->b_next; 
	    if (flag == FLUSHALL || lis_datamsg(mp->b_datap->db_type))
	    {
		lis_freemsg(rmv_msg(q,mp));
	    }

	    mp = p_next;
	}
	SPLX(psw) ;
    }
    else
	lis_flushq(q, flag) ;

}/*lis_flushband*/

/*  -------------------------------------------------------------------  */
/* lis_flushq - free messages from a queue, enabling upstream/downstream
 *	service routines if applicable. The flag is the same as for
 *	lis_flushband.
 */
void
lis_flushq(queue_t *q, int flag)
{
    mblk_t *mp, *p_next;
    int q_flag,psw;
    struct qband *qp ;

    if (q == NULL) return;

    if (!LIS_CHECK_Q_MAGIC(q)) return ;

    SPLSTR(psw);
    q_flag = q->q_flag;
    mp = q->q_first;
    q->q_first = NULL;
    q->q_last = NULL;
    q->q_count = 0;
    q->q_flag &= ~(QFULL|QWANTW|QWASFULL);
    SPLX(psw);

    while (mp != NULL)
    {
    	p_next = mp->b_next; 
	if (flag == FLUSHALL || lis_datamsg(mp->b_datap->db_type))
	{
	    LisDownCount(MSGSQD) ;		/* one fewer msg queued */
	    lis_freemsg(mp);
	}
	else
	    lis_putq(q, mp);
	mp = p_next;
    }

    SPLSTR(psw);
    for (qp = q->q_bandp; qp != NULL; qp = qp->qb_next)
    {
	qp->qb_first = NULL ;
	qp->qb_last  = NULL ;
    }
    SPLX(psw);

    SPLSTR(psw);
    q->q_flag |= q_flag & QWANTW;
    if (q->q_count <= q->q_lowat)	/* below low water?		*/
	q->q_flag |= QBACK;		/* must back enable		*/
    lis_backenable(q);
    SPLX(psw);
}/*lis_flushq*/

/*  -------------------------------------------------------------------  */
/* putctl - allocate a message block, set the type,
 *	and put it on a queue. Returns 1 on success;
 *	0 if the allocation failed or type is one of
 *	M_DATA, M_PROTO or M_PCPROTO
 */
int
lis_putctl(queue_t *q, int type, char *file_name, int line_nr)
{
    mblk_t *mp;
    if (   q == NULL
	|| !LIS_CHECK_Q_MAGIC(q)
	|| (type == M_DATA || type == M_PROTO || type == M_PCPROTO)
	|| (q->q_qinfo == NULL || q->q_qinfo->qi_putp == NULL)
	|| (mp = lis_allocb(0, BPRI_HI, file_name, line_nr)) == NULL
       )
    	   return 0;
    mp->b_datap->db_type = (uchar) type;
    (*q->q_qinfo->qi_putp)(q, mp);
    return 1;
}/*lis_putctl*/

int lis_putnextctl(queue_t *q, int type, char *file_name, int line_nr)
{
    if (q == NULL || q->q_next == NULL)
	return(0) ;				/* no can do */

    return(lis_putctl(q->q_next, type, file_name, line_nr)) ;

} /* lis_putnextctl */

/*  -------------------------------------------------------------------  */
/* lis_putctl1 - as for lis_putctl, but with a one byte parameter
 */
int
lis_putctl1(queue_t *q, int type, int param, char *file_name, int line_nr)
{
    mblk_t *mp;
    if (   q == NULL
	|| !LIS_CHECK_Q_MAGIC(q)
	|| (type == M_DATA || type == M_PROTO || type == M_PCPROTO)
	|| (q->q_qinfo == NULL || q->q_qinfo->qi_putp == NULL)
	|| (mp = lis_allocb(1, BPRI_HI, file_name, line_nr)) == NULL
       )
    	   return 0;
    mp->b_datap->db_type = (uchar) type;
    *mp->b_wptr++ = (uchar) param;
    (*(q->q_qinfo->qi_putp))(q, mp);
    return 1;
}/*lis_putctl1*/

int
lis_putnextctl1(queue_t *q, int type, int param, char *file_name, int line_nr)
{
    if (q == NULL || q->q_next == NULL)
	return(0) ;				/* no can do */

    return(lis_putctl1(q->q_next, type, param, file_name, line_nr)) ;

} /* lis_putnextctl1 */

/*  -------------------------------------------------------------------  */
/* lis_qsize - returns the number of messages on a queue
 */
int
lis_qsize(queue_t *q)
{
    mblk_t *mp;
    int rtn = 0, psw;
    if (q == NULL) return 0;

    if (!LIS_CHECK_Q_MAGIC(q)) return(0) ;

    SPLSTR(psw);
    for (mp = q->q_first; mp; mp = mp->b_next)
	rtn++;
    SPLX(psw);
    return rtn;
}/*lis_qsize*/

/*  -------------------------------------------------------------------  */
/* lis_strqget - get information about a (band of a) queue.
 *	Valid values for what are QHIWAT, QLOWAT, QMAXPSZ,
 *	QMINPSZ, QCOUNT, QFIRST, QLAST, QFLAG.
 *	Returns 0 on success.
 */
int 
lis_strqget(queue_t *q, qfields_t what, unsigned char band, long *val)
{
    struct qband *qp ;
    int		  psw;

    if (q == NULL || val == NULL) return(EINVAL) ;

    if (!LIS_CHECK_Q_MAGIC(q)) return(EINVAL) ;

    SPLSTR(psw);
    if (band > 0)
    {
	qp = find_qband(q, band) ;			/* get qband struct */
	if (qp == NULL)
	{
	    SPLX(psw) ;
	    return(EINVAL);
	}

	switch (what)
	{
	case QHIWAT:
	    *val = qp->qb_hiwat ;
	    break ;
	case QLOWAT:
	    *val = qp->qb_lowat ;
	    break ;
	case QCOUNT:
	    *val = qp->qb_count ;
	    break ;
	case QFIRST:
	    *((mblk_t **)val) = qp->qb_first ;
	    break ;
	case QLAST:
	    *((mblk_t **)val) = qp->qb_last ;
	    break ;
	case QFLAG:
	    *val = qp->qb_flag ;
	    break ;
	default:
	case QMAXPSZ:
	case QMINPSZ:
	    SPLX(psw) ;
	    return(EINVAL) ;
	}

	SPLX(psw) ;
	return(0) ;				/* success */
    }

    switch (what)
    {
    case QHIWAT:
	*val = q->q_hiwat ;
	break ;
    case QLOWAT:
	*val = q->q_lowat ;
	break ;
    case QMAXPSZ:
	*val = q->q_maxpsz ;
	break ;
    case QMINPSZ:
	*val = q->q_minpsz ;
	break ;
    case QCOUNT:
	*val = q->q_count ;
	break ;
    case QFIRST:
	*((mblk_t **)val) = q->q_first ;
	break ;
    case QLAST:
	*((mblk_t **)val) = q->q_last ;
	break ;
    case QFLAG:
	*val = q->q_flag ;
	break ;
    default:
	SPLX(psw) ;
	return(EINVAL) ;
    }

    SPLX(psw) ;
    return(0);				/* return 0 for success */

}/*lis_strqget*/

/*  -------------------------------------------------------------------  */
/* lis_strqset - change information about a (band of a) queue.
 *	Valid values for what are QHIWAT, QLOWAT, QMAXPSZ,
 *	QMINPSZ.
 *	Returns 0 on success.
 */
int
lis_strqset(queue_t *q, qfields_t what, unsigned char band, long val)
{
    struct qband *qp ;
    int		  psw;

    if (q == NULL) return(EINVAL) ;

    if (!LIS_CHECK_Q_MAGIC(q)) return(EINVAL) ;

    SPLSTR(psw);
    if (band > 0)
    {
	qp = find_qband(q, band) ;			/* get qband struct */
	if (qp == NULL)
	{
	    SPLX(psw) ;
	    return(EINVAL);
	}

	switch (what)
	{
	case QHIWAT:
	    qp->qb_hiwat = val ;
	    break ;
	case QLOWAT:
	    qp->qb_lowat = val ;
	    break ;
	case QCOUNT:
	case QFIRST:
	case QLAST:
	case QFLAG:
	    SPLX(psw) ;
	    return(EPERM) ;
	default:
	case QMAXPSZ:
	case QMINPSZ:
	    SPLX(psw) ;
	    return(EINVAL) ;
	}

	SPLX(psw) ;
	return(0) ;				/* success */
    }

    switch (what)
    {
    case QHIWAT:
	q->q_hiwat = val ;
	break ;
    case QLOWAT:
	q->q_lowat = val ;
	break ;
    case QMAXPSZ:
	q->q_maxpsz = val ;
	break ;
    case QMINPSZ:
	q->q_minpsz = val ;
	break ;
    case QCOUNT:
    case QFIRST:
    case QLAST:
    case QFLAG:
	SPLX(psw) ;
	return(EPERM) ;
    default:
	SPLX(psw) ;
	return(EINVAL) ;
    }

    SPLX(psw) ;
    return(0);				/* return 0 for success */

}/*lis_strqset*/


/*  -------------------------------------------------------------------  */
/* Allocate a new NULL-initilized queue pair
 * return NULL on failure
 */
queue_t *
lis_allocq( const char *name )
{
  queue_t *q = (queue_t*)ALLOC(sizeof(queue_t)*2);

  if (q == NULL) return(NULL) ;		/* memset does not null-check */

  LisUpCount(QUEUES) ;			/* one more queue */
  memset((void*)q,0,2*sizeof(queue_t));
  lis_mark_mem(q, name, MEM_QUEUE) ;	/* mark memory w/queue name */
  q->q_magic = Q_MAGIC ;
  (q+1)->q_magic = Q_MAGIC ;		/* can't use LIS_WR yet */
					/* now it's ok */
  q->q_flag = QREADR | QUSE ;		/* read side of queue */
  LIS_WR(q)->q_flag = QUSE ;		/* queue in use */
  return(q);
}/*lis_allocq*/

/*  -------------------------------------------------------------------  */
/* Deallocate a new queue pair
 * return NULL on failure
 */
void
lis_freeq( queue_t *q )
{
    struct qband *qp ;
    struct qband *qpx ;
    int psw;
    int i ;

    SPLSTR(psw);
    q = LIS_RD(q) ;
    for (i = 1; i <= 2; i++, q = LIS_WR(q))
    {
	if (!LIS_CHECK_Q_MAGIC(q)) continue ;

	for (qp = q->q_bandp; qp != NULL; )
	{
	    qpx = qp->qb_next ;
	    FREE(qp) ;
	    qp = qpx ;
	}
    }

    q = LIS_RD(q) ;
    (q)->q_magic   = Q_MAGIC ^ 1 ;
    (q+1)->q_magic = Q_MAGIC ^ 1 ;
    FREE(q);
    LisDownCount(QUEUES) ;			/* one fewer queue */
    SPLX(psw) ;

}/*lis_freeq*/

/*  -------------------------------------------------------------------  */
/* Insert mp2 after mp1
 * This is the same as inserting mp2 just in front of the next
 * message after mp1.
 */
void
lis_appq(queue_t *q, mblk_t *mp1, mblk_t *mp2)
{
    if (mp1 == NULL || mp2 == NULL || q == NULL)
      return;

    if (!LIS_CHECK_Q_MAGIC(q))
    {
	freemsg(mp2) ;
	return ;
    }

    lis_insq(q, mp1->b_next, mp2) ;

}/*lis_appq*/


/*  -------------------------------------------------------------------  */
/* lis_bcanput - search the stream starting from q until a service
 *	routine is found, if any, and test the found queue for
 *	flow control at a specified band. Schedule backenabling
 *	if flow controlled.
 *	For band==0 this is equivalent to canput. Returns 0 if
 *	flow controlled; 1 otherwise.
 */
int 
lis_bcanput(queue_t *q, unsigned char band)
{
    int		  psw;
    struct qband *qp ;

    if (q == NULL) return 0; /* sanity check */

    if (!LIS_CHECK_Q_MAGIC(q)) return(0) ;

    SPLSTR(psw);

    /* search the stream for the next module with a service
       procedure, and check its high water mark */
    while (   q->q_next != NULL
	   && (q->q_qinfo == NULL || q->q_qinfo->qi_srvp == NULL)
          )
	    q = q->q_next;

    /* if the queue is flow controlled, set the QWANTW flag */
    if (band > 0)
    {
	qp = find_qband(q, band) ;			/* get qband struct */
	if (qp == NULL)
	{
	    LisUpFailCount(CANPUTS) ;
	    SPLX(psw) ;
	    return(0);
	}
	if (qp->qb_flag & QB_FULL)
	{
	    qp->qb_flag |= QB_WANTW ;
	    LisUpFailCount(CANPUTS) ;
	    SPLX(psw) ;
	    return 0;
	}
    }
    else
    if (q->q_flag & QFULL)
    {
    	q->q_flag |= QWANTW;
	LisUpFailCount(CANPUTS) ;
	SPLX(psw) ;
	return 0;
    }

    LisUpCount(CANPUTS) ;
    SPLX(psw) ;
    return 1;

}/*lis_bcanput*/

/*  -------------------------------------------------------------------  */
/* lis_bcanputnext - search the stream starting from the queue after q
 * 	until a service	routine is found.  Uses lis_bcanput().
 *
 *	This routine shows up in AT&T's MP spec for SVR4.
 */
int 
lis_bcanputnext(queue_t *q, unsigned char band)
{
    if (q == NULL || !LIS_CHECK_Q_MAGIC(q)) return(0) ;

    return(lis_bcanput(q->q_next, band)) ; /* check the next queue */

}/*lis_bcanputnext*/

/*  -------------------------------------------------------------------  */
/* lis_bcanputnext_anyband - search the stream starting from q->q_next
 *	until a service	routine is found, if any, and test the found queue
 *	for flow control on any non-zero band.  This is for the S_WRBAND
 *	option processing.
 *
 *	Note, in particular, that we do not return the band number of
 *	the band with available space in it.  This actually sounds
 *	pretty useless to me.  -- DMG
 */
int 
lis_bcanputnext_anyband(queue_t *q)
{
    int		  psw;
    struct qband *qp ;

    if (   q == NULL
	|| !LIS_CHECK_Q_MAGIC(q)
	|| (q = q->q_next) == NULL
	|| !LIS_CHECK_Q_MAGIC(q)
       )
	return 0;			/* sanity check */


    SPLSTR(psw);

    /* search the stream for the next module with a service procedure */
    while (   q->q_next != NULL
	   && (q->q_qinfo == NULL || q->q_qinfo->qi_srvp == NULL)
	  )
	    q = q->q_next;

    /* Find any qband that has room in it */
    /* We even accept non-in-use qband structures, is that correct?  */
    for (qp = q->q_bandp; qp != NULL; qp = qp->qb_next)
    {
	if (!(qp->qb_flag & QB_FULL))
	    break ;
    }

    if (qp == NULL)
    {
        LisUpFailCount(CANPUTS) ;
	SPLX(psw) ;
	return(0);
    }

    LisUpCount(CANPUTS) ;
    SPLX(psw) ;
    return 1;

}/*lis_bcanputnext_anyband*/

/*  -------------------------------------------------------------------  */
/*  lis_qcountstrm
 *
 * Return the q_qcount field from the next queue.
 */
int	lis_qcountstrm(queue_t *q)
{
    int		psw;
    int		nbytes = 0 ;

    if (q == NULL || !LIS_CHECK_Q_MAGIC(q)) return(0) ;

    SPLSTR(psw);
    while (q != NULL)
    {
	if (!LIS_CHECK_Q_MAGIC(q)) break ;
	q->q_flag |= QCLOSING ;		/* mark the queue as closing */
	nbytes += q->q_count ;
	q = q->q_next ;
    }

    SPLX(psw) ;
    return(nbytes) ;

} /* lis_qcountstrm */

/*----------------------------------------------------------------------
# Local Variables:      ***
# change-log-default-name: "~/src/prj/streams/src/NOTES" ***
# End: ***
  ----------------------------------------------------------------------*/
