/************************************************************************
*                         Porting Routines                              *
*************************************************************************
*									*
* This file contains the routines with the "port_" prefix.  These	*
* routines are tailored for the USER environment.			*
*									*
* Author:	David Grothe	<dave@gcom.com>				*
*									*
************************************************************************/
/*
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 */

#ident "@(#) LiS port.c 2.3 6/21/99 21:24:02 "

#include <sys/stream.h>
#include <errno.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <time.h>

#include <sys/LiS/usrio.h>

/************************************************************************
*                        Data Structures                                *
************************************************************************/


/************************************************************************
*                         Global Storage                                *
************************************************************************/


/************************************************************************
*                           Prototypes                                  *
************************************************************************/

extern inode_t		*user_inode_of_file(int fd) ;	/* in usrio.c */
extern inode_t		*user_inode(dev_t dev) ;





/************************************************************************
*                           port_printf                                 *
*************************************************************************
*									*
* This is the printf fuction.  Return is always zero.			*
*									*
************************************************************************/
int	port_printf(char *fmt, ...)
{
    va_list args;

    va_start (args, fmt);
    vprintf (fmt, args);
    va_end (args);

    return(0) ;

} /* port_printf */

/************************************************************************
*                            port_panic                                 *
*************************************************************************
*									*
* This is the panic routine.  It is not supposed to return.		*
*									*
************************************************************************/
void	port_panic(char *msg)
{
    port_printf("\nPANIC: %s\n", msg) ;
    exit(1) ;

} /* port_panic */

/************************************************************************
*                           port_malloc                                 *
*************************************************************************
*									*
* Allocate nbytes of memory and return a pointer to it.			*
*									*
************************************************************************/
void	*port_malloc(int nbytes, int class)
{
    (void) class ;			/* memory class */

    return(malloc(nbytes)) ;

} /* port_malloc */

/************************************************************************
*                           port_free                                   *
*************************************************************************
*									*
* Free a block of memory allocated by port_malloc.			*
*									*
************************************************************************/
void	port_free(void *ptr)
{
    free(ptr) ;

} /* port_free */

/************************************************************************
*                          port_print_mem                               *
*************************************************************************
*									*
* Walk the memory link list and print out each element.			*
*									*
************************************************************************/
void	port_print_mem(void)
{
    lis_print_mem() ;

} /* port_print_mem */

/************************************************************************
*                             port_memcpy                               *
*************************************************************************
*									*
* Copy nbytes from src to dst.						*
*									*
************************************************************************/
void	port_memcpy(void *dst, void *src, int nbytes)
{
    memcpy(dst, src, nbytes) ;

} /* port_memcpy */

/************************************************************************
*                            port_verify_area                           *
*************************************************************************
*									*
* Ensure that the user area pointed to by usr_addr can be accessed	*
* for reading or writing for lgth bytes.				*
*									*
* Return negative for failure.  The absolute value of the return code	*
* is an "errno" value.  Return 0 for success.				*
*									*
************************************************************************/
int	port_verify_area(struct file *f, int rd_wr_fcn,
			 const void *usr_addr, int lgth)
{
    (void) f ;
    (void) usr_addr ;
    (void) lgth ;

    switch (rd_wr_fcn)
    {
    case VERIFY_READ:
	if (   usr_addr == NULL
	    || lgth > 0x10000
	   )
	    return(-EFAULT) ;

	return(0) ;				/* OK */
    case VERIFY_WRITE:
	if (   usr_addr == NULL
	    || lgth > 0x10000
	   )
	    return(-EFAULT) ;

	return(0) ;				/* OK */
    default:
	return(-1) ;				/* bad fcn code */
    }

} /* port_verify_area */

/************************************************************************
*                          port_memcpy_fromfs                           *
*************************************************************************
*									*
* Copy len bytes from user-space memory at ubuf into kernel space	*
* memory at kbuf.							*
*									*
* When porting to a real O/S environment, this routine will have to do	*
* more than what is coded here.						*
*									*
* Caller must guarantee that user addresses are valid before calling	*
* this routine.								*
*									*
************************************************************************/
int	port_memcpy_fromfs(struct file *f, void *kbuf,
			    const void *ubuf, int len)
{
    (void) f ;
    memcpy(kbuf, ubuf, len) ;
    return(0) ;

} /* port_memcpy_fromfs */

/************************************************************************
*                           port_memcpy_tofs                            *
*************************************************************************
*									*
* Copy len bytes from kernel-space memory at kbuf into user space	*
* memory at ubuf.							*
*									*
* When porting to a real O/S environment, this routine will have to do	*
* more than what is coded here.						*
*									*
* Caller must guarantee that user addresses are valid before calling	*
* this routine.								*
*									*
************************************************************************/
int	port_memcpy_tofs(struct file *f, const void *kbuf, void *ubuf, int len)
{
    (void) f ;
    memcpy(ubuf, kbuf, len) ;
    return(0) ;

} /* port_memcpy_tofs */

/************************************************************************
*                          port_get_fs_byte                             *
*************************************************************************
*									*
* Get a byte from user space and return it.				*
*									*
************************************************************************/
int	port_get_fs_byte(struct file *f, const void *uaddr)
{
    char	c ;

    if (port_verify_area(f, VERIFY_READ, uaddr, 1) < 0) return(0) ;

    port_memcpy_fromfs(f, &c, uaddr, 1) ;
    return( (int) c ) ;

} /* port_get_fs_byte */

/************************************************************************
*                            port_splstr                                *
*************************************************************************
*									*
* Disable interrupts to protect streams queue processing.  Return	*
* old CPU state through the pointer save_state.				*
*									*
* For user level testing, this is a dummy routine.			*
*									*
************************************************************************/
void	port_splstr(int *save_state)
{
    (void) save_state ;

} /* port_splstr */

/************************************************************************
*                             port_splx                                 *
*************************************************************************
*									*
* Restore the CPU interrupt state according to the saved state		*
* in the integer pointed to by saved_state.				*
*									*
************************************************************************/
void	port_splx(int *saved_state)
{
    (void) saved_state ;

} /* port_splx */

/************************************************************************
*                             port_spl0                                 *
*************************************************************************
*									*
* Unconditionally enable interrupts and return the previous CPU state	*
* in the integer pointed to by save_state.				*
*									*
************************************************************************/
void	port_spl0(int *save_state)
{
    (void) save_state ;

} /* port_spl0 */

/************************************************************************
*                          port_sem_P                                   *
*************************************************************************
*									*
* Perform a 'P' operation on the semaphore.				*
*									*
* This is a dummy for user-level testing.				*
*									*
************************************************************************/
int	port_sem_P(struct semaphore *sem_addr)
{
    int		flags ;

    SPLSTR(flags) ;
    if (sem_addr->sem_count > 0)
    {					/* semaphore is available */
	sem_addr->sem_count-- ;		/* acquire it */
	SPLX(flags) ;
	return(0) ;
    }

    do
    {
	SPLX(flags) ;			/* open interrupt window */
	SPLSTR(flags) ;

	port_announce_time(1) ;		/* make time appear to pass */
	lis_runqueues() ;		/* run streams scheduler */
    }
    while (sem_addr->sem_count <= 0) ;	/* exit when semaphore available */

    sem_addr->sem_count-- ;		/* acquire it */
    SPLX(flags) ;
    return(0) ;

} /* port_sem_P */

/************************************************************************
*                          port_sem_V                                   *
*************************************************************************
*									*
* Perform a 'V' operation on the semaphore.				*
*									*
* This is a dummy for user-level testing.				*
*									*
************************************************************************/
void	port_sem_V(struct semaphore *sem_addr)
{
    ++sem_addr->sem_count ;		/* give up semaphore */

} /* port_sem_V */

/************************************************************************
*                            port_sem_init                              *
*************************************************************************
*									*
* Initialize the semaphore.						*
*									*
* This is a dummy for user-level testing.				*
*									*
************************************************************************/
void	port_sem_init(struct semaphore *sem_addr, int counter)
{
    sem_addr->sem_count = counter ;

} /* port_sem_init */

/************************************************************************
*                            port_sem_destroy                           *
*************************************************************************
*									*
* De-initialize the semaphore.						*
*									*
* This is a dummy for user-level testing.				*
*									*
************************************************************************/
void	port_sem_destroy(struct semaphore *sem_addr)
{
    memset(sem_addr, 0, sizeof(*sem_addr)) ;

} /* port_sem_destroy */

/************************************************************************
*                          port_get_uid                                 *
*************************************************************************
*									*
* Get user id.								*
*									*
************************************************************************/
int	 port_get_uid(struct file *f)
{
    (void) f ;
    return(0 /*204*/) ;

} /* port_get_uid */

/************************************************************************
*                           port_get_gid                                *
*************************************************************************
*									*
* Get group id.								*
*									*
************************************************************************/
int	 port_get_gid(struct file *f)
{
    (void) f ;
    return(0 /*145*/) ;

} /* port_get_gid */

/************************************************************************
*                          port_get_euid                                *
*************************************************************************
*									*
* Get effective user id.						*
*									*
************************************************************************/
int	 port_get_euid(struct file *f)
{
    (void) f ;
    return(0 /*205*/) ;

} /* port_get_euid */

/************************************************************************
*                           port_get_egid                               *
*************************************************************************
*									*
* Get effective group id.						*
*									*
************************************************************************/
int	 port_get_egid(struct file *f)
{
    (void) f ;
    return(0 /*146*/) ;

} /* port_get_egid */

/************************************************************************
*                             port_get_pgrp                             *
*************************************************************************
*									*
* Get process group id.							*
*									*
************************************************************************/
int	 port_get_pgrp(struct file *f)
{
    (void) f ;
    return(33) ;

} /* port_get_pgrp */

/************************************************************************
*                              port_get_pid                             *
*************************************************************************
*									*
* Get process id.							*
*									*
************************************************************************/
int	 port_get_pid(struct file *f)
{
    (void) f ;
    return(19) ;

} /* port_get_pid */

/************************************************************************
*                             port_suser                                *
*************************************************************************
*									*
* Return true if we are running as the super user, false otherwise.	*
*									*
************************************************************************/
int	port_suser(struct file *f)
{
    return(port_get_uid(f) == 0) ;

} /* port_suser */

/************************************************************************
*                           port_kill_proc                              *
*************************************************************************
*									*
* Send the signal 'sig' to the process 'pid'.  'priv' is true if the	*
* user is super-user.  Return is negative error number or zero for	*
* success.								*
*									*
************************************************************************/
int	port_kill_proc(int pid, int sig, int priv)
{
    (void) pid ;			/* compiler happiness */
    (void) sig ;			/* compiler happiness */
    (void) priv ;			/* compiler happiness */
    return(0) ;

} /* port_kill_proc */

/************************************************************************
*                            port_kill_pg                               *
*************************************************************************
*									*
* Send the signal 'sig' to the process group 'pgrp'.  'priv' is true	*
* if the user is super-user.  Return is negative error number or	*
* zero for success.							*
*									*
************************************************************************/
int	port_kill_pg(int pgrp, int sig, int priv)
{
    (void) pgrp ;			/* compiler happiness */
    (void) sig ;			/* compiler happiness */
    (void) priv ;			/* compiler happiness */
    return(0) ;

} /* port_kill_pg */

/************************************************************************
*                         port_fd_to_str				*
*************************************************************************
*									*
* Given in integer file desriptor, fd, return a pointer to the stream	*
* head structure associated with the file.				*
*									*
************************************************************************/
stdata_t	*port_fd_to_str(int fd)
{
    inode_t	*ip ;

    ip = user_inode_of_file(fd) ;
    if (ip == NULL) return(NULL) ;

    return(ip->i_str) ;		/* return ptr to sdtata struct from inode */

} /* port_fd_to_str */

/************************************************************************
*                            port_session                               *
*************************************************************************
*									*
* Return the session number of the current process.  I do not understand*
* what this is supposed to be and it may be superfluous.		*
*									*
************************************************************************/
int	port_session(struct file *f)
{
    return(port_get_pid(f)) ;

} /* port_session */

/************************************************************************
*                         port_setqsched                                *
*************************************************************************
*									*
* This routine is supposed to schedule the queues to be run.  		*
* We do nothing.  The LiS entry point routines all call lis_runqueues	*
* on their way out which causes lis_run_queues to be called to actually	*
* run the queues.							*
*									*
************************************************************************/
void	port_setqsched(void)
{
#if 0					/* do nothing */
    lis_run_queues() ;
#endif

} /* port_setqsched */

/************************************************************************
*                            port_pick_inode                            *
*************************************************************************
*									*
* This routine is called from stropen().				*
*									*
* From the two inodes passed in, choose one of them to return to the	*
* caller and "dispose" of the other one.				*
*									*
* 'old' is the inode passed into stropen by sys_open, 'new' is an	*
* inode with the same i_rdev that was allocated to a previously opened	*
* stream (so it's really the "older" of the two).			*
*									*
* Alternatively, 'old' is the inode passed to stropen for a clone open	*
* and 'new' is NULL.  In this case we need to allocate a new inode,	*
* transfer some attributes into it and return a pointer to it.  We are	*
* done with the 'old' inode since it was the one that pointed to the	*
* /dev/clone and the new one points to the actual stream device opened.	*
*									*
************************************************************************/
struct inode *port_pick_inode(struct inode *old, struct inode *new, dev_t dev)
{
    file_t	*fp ;
    dir_t	*dp ;
    char	 name_buf[MAXNAMLEN];
    int		 rslt ;
    int		 cloned = 0 ;

    if (new == NULL)				/* no new inode */
    {						/* may need to make one */
	if (old != NULL && old->i_dev == dev)
	    return(old) ;			/* use old one */

	new = user_inode(dev) ;			/* ptr to inode */
	if (new == NULL)			/* couldn't allocate one */
	    goto free_old ;			/* old is wrong dev */

	ULOCK_INO(new) ;
	if (old != NULL)			/* new is cloned */
	{
	    cloned       = 1 ;
	    new->i_str   = old->i_str ;		/* xfer stream ptr */
	    new->i_mode  = old->i_mode ;	/* inherit mode */
	    new->i_ftime = old->i_ftime;
	    new->i_ctime = old->i_ctime;
	    new->i_atime = old->i_atime;
	    new->i_mtime = old->i_mtime;
	    goto free_old ;
	}
	else
	{
	    new->i_ftime = time(NULL);
	    new->i_ctime = new->i_ftime;
	    new->i_atime = new->i_ftime;
	    new->i_mtime = new->i_ftime;
	}
    }

    /*
     * We have both old and new.  Use the "new" one and free up the
     * "old" one.
     */

free_old:					/* free old and return new */
    if (old != NULL)
	fp = user_file_of_inode(old) ;
    else
	fp = user_alloc_file() ;

    if (fp != NULL)
    {						/* needs a name */
	dp = user_search_dir_inode(new) ;
	if (dp != NULL)
	    strncpy(name_buf, dp->d_name, sizeof(name_buf)) ;
	else
	    sprintf(name_buf, "%s%s",
		    lis_strm_name(new->i_str),
		    lis_maj_min_name(new->i_str)) ;

	if (fp->f_name != NULL)			/* free any old name */
	    FREE(fp->f_name) ;

	fp->f_name = ALLOCF(strlen(name_buf) + 1, "clone-name ") ;
	if (fp->f_name == NULL)
	    fp->f_name = "Anonymous-File" ;
	else
	    strcpy(fp->f_name, name_buf) ;	/* save new name */

	if (dp == NULL)				/* no directory entry */
	{
	    rslt = user_add_dir_entry(name_buf, new->i_mode, dev) ;
	    if (rslt != 0)
	    {
		printk("port_pick_inode: error %d adding '%s' to directory\n",
			rslt, name_buf) ;
	    }
	}

	if ( (dp = user_search_dir(name_buf)) == NULL)
	{
	    printk("port_pick_inode: could not find '%s' in directory\n",
		    name_buf) ;
	}
	else
	{
	    new->i_inum = dp->d_inum  ;		/* assign inode nr */
	    dp->d_refcnt++ ;
	}
    }

    if (old != NULL)
    {
	old->i_use-- ;
	old->i_str = NULL ;
	if (old->i_use == 0)
	    lis_mark_mem(old, "unused inode: ", major(old->i_dev)) ;
    }

    return(new) ;				/* use new one */

} /* port_pick_inode */

/************************************************************************
*                         port_select_wakeup                            *
*************************************************************************
*									*
* A dummy for select mechanism (which we don't implement).		*
*									*
************************************************************************/
void port_select_wakeup(stdata_t *hd)
{
} /* port_select_wakeup */
